"""Simple implementation of ULIDs."""

from __future__ import annotations

import abc
import array
import binascii
import datetime
import os
import time
import typing as T
import uuid
from datetime import UTC

import typing_extensions as TX

from unipercept.types import Buffer, Datetime, Primitive

__all__ = ["ULID", "Randomness", "Timestamp", "base32_decode", "base32_encode"]


class _MemView:
    """
    Wraps a buffer object, typically :class:`bytes`, with a :class:`memoryview` and provides easy
    type comparisons and conversions between presentation formats.
    """

    __slots__ = ["memory"]

    def __init__(self, buffer: Buffer) -> None:
        self.memory = memoryview(buffer)

    @TX.override
    def __eq__(self, other: MemViewPrimitive) -> bool:
        if isinstance(other, _MemView):
            return self.memory == other.memory
        if isinstance(other, (bytes, bytearray, memoryview)):
            return self.memory == other
        if isinstance(other, int):
            return self.int == other
        if isinstance(other, float):
            return self.float == other
        if isinstance(other, str):
            return self.str == other
        return NotImplemented

    @TX.override
    def __ne__(self, other: MemViewPrimitive) -> bool:
        if isinstance(other, _MemView):
            return self.memory != other.memory
        if isinstance(other, (bytes, bytearray, memoryview)):
            return self.memory != other
        if isinstance(other, int):
            return self.int != other
        if isinstance(other, float):
            return self.float != other
        if isinstance(other, str):
            return self.str != other
        return NotImplemented

    def __lt__(self, other: MemViewPrimitive) -> bool:
        if isinstance(other, _MemView):
            return self.int < other.int
        if isinstance(other, (bytes, bytearray)):
            return self.bytes < other
        if isinstance(other, memoryview):
            return self.bytes < other.tobytes()
        if isinstance(other, int):
            return self.int < other
        if isinstance(other, float):
            return self.float < other
        if isinstance(other, str):
            return self.str < other
        return NotImplemented

    def __gt__(self, other: MemViewPrimitive) -> bool:
        if isinstance(other, _MemView):
            return self.int > other.int
        if isinstance(other, (bytes, bytearray)):
            return self.bytes > other
        if isinstance(other, memoryview):
            return self.bytes > other.tobytes()
        if isinstance(other, int):
            return self.int > other
        if isinstance(other, float):
            return self.float > other
        if isinstance(other, str):
            return self.str > other
        return NotImplemented

    def __le__(self, other: MemViewPrimitive) -> bool:
        if isinstance(other, _MemView):
            return self.int <= other.int
        if isinstance(other, (bytes, bytearray)):
            return self.bytes <= other
        if isinstance(other, memoryview):
            return self.bytes <= other.tobytes()
        if isinstance(other, int):
            return self.int <= other
        if isinstance(other, float):
            return self.float <= other
        if isinstance(other, str):
            return self.str <= other
        return NotImplemented

    def __ge__(self, other: MemViewPrimitive) -> bool:
        if isinstance(other, _MemView):
            return self.int >= other.int
        if isinstance(other, (bytes, bytearray)):
            return self.bytes >= other
        if isinstance(other, memoryview):
            return self.bytes >= other.tobytes()
        if isinstance(other, int):
            return self.int >= other
        if isinstance(other, float):
            return self.float >= other
        if isinstance(other, str):
            return self.str >= other
        return NotImplemented

    @TX.override
    def __hash__(self) -> int:
        return hash(self.memory)

    def __bytes__(self) -> bytes:
        return self.bytes

    def __float__(self) -> float:
        return self.float

    def __int__(self) -> int:
        return self.int

    def __index__(self) -> int:
        return self.int

    @TX.override
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({str(self)!r})>"

    @TX.override
    def __str__(self) -> str:
        return self.str

    @TX.override
    def __getstate__(self) -> str:
        return self.str

    def __setstate__(self, state: str) -> None:
        self.memory = memoryview(base32_decode(state))

    @property
    def bin(self) -> str:
        return bin(self.int)

    @property
    def bytes(self) -> bytes:
        return self.memory.tobytes()

    @property
    def float(self) -> float:
        return float(self.int)

    @property
    def hex(self) -> str:
        return "0x" + binascii.hexlify(self.bytes).decode()

    @property
    def int(self) -> int:
        return int.from_bytes(self.memory, byteorder="big")

    @property
    def oct(self) -> str:
        return oct(self.int)

    @property
    @abc.abstractmethod
    def str(self) -> str:
        return NotImplemented


MemViewPrimitive = _MemView | Primitive


class Timestamp(_MemView):
    """
    Represents the timestamp portion of a ULID.

    * Unix time (time since epoch) in milliseconds.
    * First 48 bits of ULID when in binary format.
    * First 10 characters of ULID when in string format.
    """

    __slots__ = _MemView.__slots__

    @classmethod
    def generate(cls) -> T.Self:
        return cls.create(int(time.time() * 1000).to_bytes(6, "big"))

    @classmethod
    def create(cls, timestamp: T.Self | Primitive | Datetime | ULID) -> T.Self:
        """
        Create a new :class:`ULID` instance using a timestamp value of a supported type.

        The following types are supported for timestamp values:

        * :class:`datetime.datetime`
        * :class:`int`
        * :class:`float`
        * :class:`str`
        * :class:`memoryview`
        * :class:`Timestamp`
        * :class:`ULID`
        * :class:`bytes`
        * :class:`bytearray`

        Parameters
        ----------
        timestamp : datetime, int, float, str, memoryview, Timestamp, ULID, bytes, bytearray
            Unix timestamp in seconds

        Returns
        -------
        :class:`ULID`
            ULID using given timestamp and new randomness

        Raises
        ------
        ValueError
            When the value is an unsupported type
        ValueError
            When the value is a string and cannot be Base32 decoded
        ValueError
            When the value is or was converted to something 48 bits
        """
        if isinstance(timestamp, datetime.datetime):
            timestamp = timestamp.timestamp()
        if isinstance(timestamp, (int, float)):
            timestamp = int(timestamp * 1000.0).to_bytes(6, byteorder="big")
        elif isinstance(timestamp, str):
            timestamp = _base32_decode_timestamp(timestamp)
        elif isinstance(timestamp, memoryview):
            timestamp = timestamp.tobytes()
        elif isinstance(timestamp, cls):
            timestamp = timestamp.bytes
        elif isinstance(timestamp, ULID):
            timestamp = timestamp.timestamp().bytes

        if not isinstance(timestamp, (bytes, bytearray)):
            msg = (
                "Expected datetime, int, float, str, memoryview, Timestamp, ULID, "
                "bytes, or bytearray; got %s"
            )
            raise ValueError(msg, type(timestamp).__name__)

        length = len(timestamp)
        if length != 6:
            msg = f"Expects timestamp to be 48 bits; got {length} bytes"
            raise ValueError(msg)

        return cls(timestamp)

    @property
    @TX.override
    def str(self) -> str:
        """
        Computes the string value of the timestamp from the underlying :class:`memoryview` in Base32 encoding.

        Parameters
        ----------
        None

        Returns
        -------
        str
            Timestamp in Base32 string form.

        Raises
        ------
        ValueError
            If underlying :class:`memoryview` cannot be encoded.
        """
        return _base32_encode_timestamp(self.memory)

    @property
    def timestamp(self) -> float:
        """
        Computes the Unix time (seconds since epoch) from its :class:`memoryview`.

        Returns
        -------
        float
            Timestamp in Unix time (seconds since epoch) form.
        """
        return self.int / 1000.0

    @property
    def datetime(self) -> Datetime:
        """
        Creates a :class:`datetime.datetime` instance (assumes UTC) from the Unix time value of the timestamp
        with millisecond precision.

        Returns
        -------
        :class:`datetime.datetime`
            Timestamp in datetime form.
        """
        milli = self.int
        micro = milli % 1000 * 1000
        sec = milli // 1000.0
        timezone = UTC

        return datetime.datetime.utcfromtimestamp(sec).replace(
            microsecond=micro, tzinfo=timezone
        )

    @property
    @TX.override
    def str(self) -> str:
        return _base32_encode_timestamp(self.memory)


class Randomness(_MemView):
    """
    Represents the randomness portion of a ULID.

    * Cryptographically secure random values.
    * Last 80 bits of ULID when in binary format.
    * Last 16 characters of ULID when in string format.
    """

    __slots__ = _MemView.__slots__

    @classmethod
    def generate(cls) -> T.Self:
        return cls.create(os.urandom(10))

    @property
    @TX.override
    def str(self) -> str:
        """
        Computes the string value of the randomness from the underlying :class:`memoryview` in Base32 encoding.

        Returns
        -------
        str
            Timestamp in Base32 string form.

        Raises
        ------
        ValueError
            If underlying :class:`memoryview` cannot be encoded
        """
        return _base32_encode_timestamp(self.memory)

    @classmethod
    def create(cls, randomness: T.Self | Primitive) -> T.Self:
        """
        Create a new :class:`Randomness` instance using the given randomness value of a supported type.

        The following types are supported for randomness values:

        * :class:`int`
        * :class:`float`
        * :class:`str`
        * :class:`memoryview`
        * :class:`Randomness`
        * :class:`ULID`
        * :class:`bytes`
        * :class:`bytearray`

        Parameters
        ----------
        randomness : int, float, str, memoryview, Randomness, ULID, bytes, bytearray
            Random bytes

        Returns
        -------
        :class:`ULID`
            ULID using new timestamp and given randomness

        Raises
        ------
        ValueError
            When the value is an unsupported type
        ValueError
            When the value is a string and cannot be Base32 decoded
        ValueError
            When the value is or was converted to something 80 bits
        """
        if isinstance(randomness, (int, float)):
            randomness = int(randomness).to_bytes(10, byteorder="big")
        elif isinstance(randomness, str):
            randomness = _base32_decode_randomness(randomness)
        elif isinstance(randomness, memoryview):
            randomness = randomness.tobytes()
        elif isinstance(randomness, Randomness):
            randomness = randomness.bytes
        elif isinstance(randomness, ULID):
            randomness = randomness.randomness().bytes

        if not isinstance(randomness, (bytes, bytearray)):
            msg = (
                "Expected int, float, str, memoryview, Randomness, ULID, "
                f"bytes, or bytearray; got {type(randomness).__name__}"
            )
            raise ValueError(msg)

        length = len(randomness)
        if length != 10:
            msg = "Expected randomness to be 80 bits; got %d bytes"
            raise ValueError(msg, length)

        return cls(randomness)

    @property
    @TX.override
    def str(self) -> str:
        return _base32_encode_timestamp(self.memory)


class ULID(_MemView):
    """
    Represents a ULID.

    * 128 bits in binary format.
    * 26 characters in string format.
    * 16 octets.
    * Network byte order, big-endian, most significant bit first.
    """

    __slots__ = _MemView.__slots__

    @classmethod
    def generate(cls) -> T.Self:
        """
        Generate a new :class:`ULID` instance using the current time and cryptographically secure
        randomness.

        Returns
        -------
        :class:`ULID`
            New ULID instance
        """
        timestamp = Timestamp.generate()
        randomness = Randomness.generate()

        return cls.create(timestamp, randomness)

    @classmethod
    def create(
        cls,
        timestamp: Timestamp | Primitive | Datetime | ULID,
        randomness: Randomness | Primitive,
    ) -> T.Self:
        if not isinstance(timestamp, Timestamp):
            timestamp = Timestamp.create(timestamp)
        if not isinstance(randomness, Randomness):
            randomness = Randomness.create(randomness)

        ulid = timestamp.bytes + randomness.bytes

        return cls(ulid)

    @property
    @TX.override
    def str(self) -> str:
        """
        Computes the string value of the ULID from its :class:`memoryview` in Base32 encoding.

        Returns
        -------
        str
            ULID in Base32 string form.

        Raises
        ------
        ValueError
            If underlying :class:`memoryview` cannot be encoded.
        """
        return _base32_encode_ulid(self.memory)

    def timestamp(self) -> Timestamp:
        """
        Creates a :class:`Timestamp` instance that maps to the first 48 bits of this ULID.

        Returns
        -------
        :class:`Timestamp`
            Timestamp from first 48 bits.
        """
        return Timestamp(self.memory[:6])

    def randomness(self) -> Randomness:
        """
        Creates a :class:`Randomness` instance that maps to the last 80 bits of this ULID.

        Returns
        -------
        :class:`Timestamp`
            Timestamp from first 48 bits.
        """
        return Randomness(self.memory[6:])

    @property
    def uuid(self) -> uuid.UUID:
        """
        Creates a :class:`uuid.UUID` instance of the ULID from its :class:`bytes` representation.

        Returns
        -------
        :class:`uuid.UUID`
            UUIDv4 from the ULID bytes
        """
        return uuid.UUID(bytes=self.bytes)

    @property
    @TX.override
    def str(self) -> str:
        return _base32_encode_ulid(self.memory)


BASE32_CHARSET = "0123456789ABCDEFGHJKMNPQRSTVWXYZ"  # NOTE: Exclude 'I', 'L', 'O', 'U'
BASE32_CHARMAP = array.array(
    "B",
    (
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0x00,
        0x01,
        0x02,
        0x03,
        0x04,
        0x05,
        0x06,
        0x07,
        0x08,
        0x09,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0x0A,
        0x0B,
        0x0C,
        0x0D,
        0x0E,
        0x0F,
        0x10,
        0x11,
        0x01,
        0x12,
        0x13,
        0x01,
        0x14,
        0x15,
        0x00,
        0x16,
        0x17,
        0x18,
        0x19,
        0x1A,
        0xFF,
        0x1B,
        0x1C,
        0x1D,
        0x1E,
        0x1F,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0x0A,
        0x0B,
        0x0C,
        0x0D,
        0x0E,
        0x0F,
        0x10,
        0x11,
        0x01,
        0x12,
        0x13,
        0x01,
        0x14,
        0x15,
        0x00,
        0x16,
        0x17,
        0x18,
        0x19,
        0x1A,
        0xFF,
        0x1B,
        0x1C,
        0x1D,
        0x1E,
        0x1F,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
    ),
)


def base32_encode(value: Buffer) -> str:
    length = len(value)

    # Order here is based on assumed hot path.
    if length == 16:
        return _base32_encode_ulid(value)
    if length == 6:
        return _base32_encode_timestamp(value)
    if length == 10:
        return _base32_encode_timestamp(value)

    msg = "Expects bytes in sizes of 6, 10, or 16; got %d"
    raise ValueError(msg, length)


def _base32_encode_ulid(value: Buffer) -> str:
    length = len(value)
    if length != 16:
        msg = "Expects 16 bytes for timestamp + randomness; got %d"
        raise ValueError(msg, length)

    encoding = BASE32_CHARSET

    return (
        encoding[(value[0] & 224) >> 5]
        + encoding[value[0] & 31]
        + encoding[(value[1] & 248) >> 3]
        + encoding[((value[1] & 7) << 2) | ((value[2] & 192) >> 6)]
        + encoding[((value[2] & 62) >> 1)]
        + encoding[((value[2] & 1) << 4) | ((value[3] & 240) >> 4)]
        + encoding[((value[3] & 15) << 1) | ((value[4] & 128) >> 7)]
        + encoding[(value[4] & 124) >> 2]
        + encoding[((value[4] & 3) << 3) | ((value[5] & 224) >> 5)]
        + encoding[value[5] & 31]
        + encoding[(value[6] & 248) >> 3]
        + encoding[((value[6] & 7) << 2) | ((value[7] & 192) >> 6)]
        + encoding[(value[7] & 62) >> 1]
        + encoding[((value[7] & 1) << 4) | ((value[8] & 240) >> 4)]
        + encoding[((value[8] & 15) << 1) | ((value[9] & 128) >> 7)]
        + encoding[(value[9] & 124) >> 2]
        + encoding[((value[9] & 3) << 3) | ((value[10] & 224) >> 5)]
        + encoding[value[10] & 31]
        + encoding[(value[11] & 248) >> 3]
        + encoding[((value[11] & 7) << 2) | ((value[12] & 192) >> 6)]
        + encoding[(value[12] & 62) >> 1]
        + encoding[((value[12] & 1) << 4) | ((value[13] & 240) >> 4)]
        + encoding[((value[13] & 15) << 1) | ((value[14] & 128) >> 7)]
        + encoding[(value[14] & 124) >> 2]
        + encoding[((value[14] & 3) << 3) | ((value[15] & 224) >> 5)]
        + encoding[value[15] & 31]
    )


def _base32_encode_timestamp(timestamp: Buffer) -> str:
    length = len(timestamp)
    if length != 6:
        msg = "Expects 6 bytes for timestamp; got %d"
        raise ValueError(msg, length)

    encoding = BASE32_CHARSET

    return (
        encoding[(timestamp[0] & 224) >> 5]
        + encoding[timestamp[0] & 31]
        + encoding[(timestamp[1] & 248) >> 3]
        + encoding[((timestamp[1] & 7) << 2) | ((timestamp[2] & 192) >> 6)]
        + encoding[((timestamp[2] & 62) >> 1)]
        + encoding[((timestamp[2] & 1) << 4) | ((timestamp[3] & 240) >> 4)]
        + encoding[((timestamp[3] & 15) << 1) | ((timestamp[4] & 128) >> 7)]
        + encoding[(timestamp[4] & 124) >> 2]
        + encoding[((timestamp[4] & 3) << 3) | ((timestamp[5] & 224) >> 5)]
        + encoding[timestamp[5] & 31]
    )


def _base32_encode_timestamp(randomness: Buffer) -> str:
    length = len(randomness)
    if length != 10:
        msg = "Expects 10 bytes for randomness; got %d"
        raise ValueError(msg, length)

    encoding = BASE32_CHARSET

    return (
        encoding[(randomness[0] & 248) >> 3]
        + encoding[((randomness[0] & 7) << 2) | ((randomness[1] & 192) >> 6)]
        + encoding[(randomness[1] & 62) >> 1]
        + encoding[((randomness[1] & 1) << 4) | ((randomness[2] & 240) >> 4)]
        + encoding[((randomness[2] & 15) << 1) | ((randomness[3] & 128) >> 7)]
        + encoding[(randomness[3] & 124) >> 2]
        + encoding[((randomness[3] & 3) << 3) | ((randomness[4] & 224) >> 5)]
        + encoding[randomness[4] & 31]
        + encoding[(randomness[5] & 248) >> 3]
        + encoding[((randomness[5] & 7) << 2) | ((randomness[6] & 192) >> 6)]
        + encoding[(randomness[6] & 62) >> 1]
        + encoding[((randomness[6] & 1) << 4) | ((randomness[7] & 240) >> 4)]
        + encoding[((randomness[7] & 15) << 1) | ((randomness[8] & 128) >> 7)]
        + encoding[(randomness[8] & 124) >> 2]
        + encoding[((randomness[8] & 3) << 3) | ((randomness[9] & 224) >> 5)]
        + encoding[randomness[9] & 31]
    )


def base32_decode(value: str) -> bytes:
    length = len(value)

    # Order here is based on assumed hot path.
    if length == 26:
        return _base32_decode_ulid(value)
    if length == 10:
        return _base32_decode_timestamp(value)
    if length == 16:
        return _base32_decode_randomness(value)

    msg = "Expects string in lengths of 10, 16, or 26; got %d"
    raise ValueError(msg, length)


def _base32_decode_ulid(value: str) -> bytes:
    encoded = _base32_as_bytes(value, 26)

    decoding = BASE32_CHARMAP

    return bytes(
        (
            ((decoding[encoded[0]] << 5) | decoding[encoded[1]]) & 0xFF,
            ((decoding[encoded[2]] << 3) | (decoding[encoded[3]] >> 2)) & 0xFF,
            (
                (decoding[encoded[3]] << 6)
                | (decoding[encoded[4]] << 1)
                | (decoding[encoded[5]] >> 4)
            )
            & 0xFF,
            ((decoding[encoded[5]] << 4) | (decoding[encoded[6]] >> 1)) & 0xFF,
            (
                (decoding[encoded[6]] << 7)
                | (decoding[encoded[7]] << 2)
                | (decoding[encoded[8]] >> 3)
            )
            & 0xFF,
            ((decoding[encoded[8]] << 5) | (decoding[encoded[9]])) & 0xFF,
            ((decoding[encoded[10]] << 3) | (decoding[encoded[11]] >> 2)) & 0xFF,
            (
                (decoding[encoded[11]] << 6)
                | (decoding[encoded[12]] << 1)
                | (decoding[encoded[13]] >> 4)
            )
            & 0xFF,
            ((decoding[encoded[13]] << 4) | (decoding[encoded[14]] >> 1)) & 0xFF,
            (
                (decoding[encoded[14]] << 7)
                | (decoding[encoded[15]] << 2)
                | (decoding[encoded[16]] >> 3)
            )
            & 0xFF,
            ((decoding[encoded[16]] << 5) | (decoding[encoded[17]])) & 0xFF,
            ((decoding[encoded[18]] << 3) | (decoding[encoded[19]] >> 2)) & 0xFF,
            (
                (decoding[encoded[19]] << 6)
                | (decoding[encoded[20]] << 1)
                | (decoding[encoded[21]] >> 4)
            )
            & 0xFF,
            ((decoding[encoded[21]] << 4) | (decoding[encoded[22]] >> 1)) & 0xFF,
            (
                (decoding[encoded[22]] << 7)
                | (decoding[encoded[23]] << 2)
                | (decoding[encoded[24]] >> 3)
            )
            & 0xFF,
            ((decoding[encoded[24]] << 5) | (decoding[encoded[25]])) & 0xFF,
        )
    )


def _base32_decode_timestamp(timestamp: str) -> bytes:
    encoded = _base32_as_bytes(timestamp, 10)

    decoding = BASE32_CHARMAP

    return bytes(
        (
            ((decoding[encoded[0]] << 5) | decoding[encoded[1]]) & 0xFF,
            ((decoding[encoded[2]] << 3) | (decoding[encoded[3]] >> 2)) & 0xFF,
            (
                (decoding[encoded[3]] << 6)
                | (decoding[encoded[4]] << 1)
                | (decoding[encoded[5]] >> 4)
            )
            & 0xFF,
            ((decoding[encoded[5]] << 4) | (decoding[encoded[6]] >> 1)) & 0xFF,
            (
                (decoding[encoded[6]] << 7)
                | (decoding[encoded[7]] << 2)
                | (decoding[encoded[8]] >> 3)
            )
            & 0xFF,
            ((decoding[encoded[8]] << 5) | (decoding[encoded[9]])) & 0xFF,
        )
    )


def _base32_decode_randomness(randomness: str) -> bytes:
    encoded = _base32_as_bytes(randomness, 16)

    decoding = BASE32_CHARMAP

    return bytes(
        (
            ((decoding[encoded[0]] << 3) | (decoding[encoded[1]] >> 2)) & 0xFF,
            (
                (decoding[encoded[1]] << 6)
                | (decoding[encoded[2]] << 1)
                | (decoding[encoded[3]] >> 4)
            )
            & 0xFF,
            ((decoding[encoded[3]] << 4) | (decoding[encoded[4]] >> 1)) & 0xFF,
            (
                (decoding[encoded[4]] << 7)
                | (decoding[encoded[5]] << 2)
                | (decoding[encoded[6]] >> 3)
            )
            & 0xFF,
            ((decoding[encoded[6]] << 5) | (decoding[encoded[7]])) & 0xFF,
            ((decoding[encoded[8]] << 3) | (decoding[encoded[9]] >> 2)) & 0xFF,
            (
                (decoding[encoded[9]] << 6)
                | (decoding[encoded[10]] << 1)
                | (decoding[encoded[11]] >> 4)
            )
            & 0xFF,
            ((decoding[encoded[11]] << 4) | (decoding[encoded[12]] >> 1)) & 0xFF,
            (
                (decoding[encoded[12]] << 7)
                | (decoding[encoded[13]] << 2)
                | (decoding[encoded[14]] >> 3)
            )
            & 0xFF,
            ((decoding[encoded[14]] << 5) | (decoding[encoded[15]])) & 0xFF,
        )
    )


def _base32_as_bytes(value: str, expected_length: int) -> bytes:
    """
    Convert the given string to bytes and validate it is within the Base32 character set.

    Parameters
    ----------
    value : str
        String to convert to bytes
    expected_length : int
        Expected length of the input string

    Returns
    -------
    bytes
        Value converted to bytes.
    """
    length = len(value)
    if length != expected_length:
        msg = "Expects %d characters for decoding; got %d"
        raise ValueError(msg, expected_length, length)

    try:
        encoded = value.encode("ascii")
    except UnicodeEncodeError as ex:
        msg = "Expects value that can be encoded in ASCII charset: %s"
        raise ValueError(msg, ex) from ex

    decoding = BASE32_CHARMAP

    # Confirm all bytes are valid Base32 decode characters.
    # Note: ASCII encoding handles the out of range checking for us.
    for byte in encoded:
        if decoding[byte] > 31:
            msg = f"Non-base32 character found: {chr(byte)!r}"
            raise ValueError(msg)

    # Confirm most significant bit on timestamp value is limited so it can be stored in 128-bits.
    if length in (10, 26):
        msb = decoding[encoded[0]]
        if msb > 7:
            msg = (
                "Timestamp value too large and will overflow 128-bits. "
                'Must be between b"0" and b"7"'
            )
            raise ValueError(msg)

    return encoded



from __future__ import annotations

from collections.abc import Callable
from functools import update_wrapper
from types import MethodType
from typing import (
    Any,
    Concatenate,
    Final,
    ParamSpec,
    TypeVar,
    cast,
    overload,
    override,
)

from unipercept.utils.missing import MissingValue

__all__ = ["immutable", "objectmagic"]

_P = ParamSpec("_P")
_T = TypeVar("_T", bound=object)
_R = TypeVar("_R", covariant=True)


class objectmagic[T: object, **P, R]:
    """
    A variant on classmethod that can only be called on instances of a class.
    Useful for wrapping magic methods while still preserving the ability to call such a method statically.
    """

    @property
    def __func__(self) -> Callable[Concatenate[_T, _P], _R]: ...

    def __init__(self, fn: Callable[Concatenate[_T, _P], _R]) -> None:
        self.fn = fn
        self.name = fn.__name__
        self.owner = None

        update_wrapper(self, fn)  # type: ignore

    def __set_name__(self, owner: type[_T], name: str) -> None:
        self.name = name
        self.owner = owner

    @overload
    def __get__(self, obj: None, *args, **kwargs) -> Callable[_P, _R]: ...

    @overload
    def __get__(
        self, obj: _T, *args, **kwargs
    ) -> Callable[Concatenate[_T, _P], _R]: ...

    def __get__(
        self, obj: _T | None, *args, **kwargs
    ) -> Callable[Concatenate[_T, _P], _R] | Callable[_P, _R]:
        if obj is None:
            return self.fn  # (obj_a, obj_b)
        method = MethodType(self.fn, obj)  # type: ignore
        return cast(Callable[P, R], method)  # (obj_b)


_V = TypeVar("_V", covariant=True)
_NA: Final = MissingValue("NA")


class immutable[V]:
    """A value that can only be set once and not changed afterwards."""

    def __init__(self, value: _V | MissingValue = _NA) -> None:
        self.value: _V = cast(V, value)
        self.name: str | None = None

    def __get__(self, obj: object, cls: type[object]) -> _V:
        res = self.value
        if _NA.is_value(res):
            return res
        msg = f"{self.name} is not set"
        raise AttributeError(msg)

    def __set_name__(self, owner: type[object], name: str) -> None:
        self.name = name

    def __set__(self, obj: object, value: Any, /) -> None:
        if self.name is not None:
            msg = f"{self.name} is read-only"
            raise AttributeError(msg)
        self.value = value

    @override
    def __str__(self) -> str:
        return str(self.value)

    @override
    def __repr__(self) -> str:
        return f"<immutable {type(self.value)} {self!s}>"


class private[V]:
    """
    A property that can only be accessed by the class that owns it. It is not accessible by subclasses.
    This is useful for properties that are only meant to be applicable by the class itself, like a version string.

    Note that this is not a security feature, and is not meant to be used as such.
    """

    value: _V | MissingValue
    owner: type | None
    name: str | None

    def __init__(
        self, value: _V | MissingValue = _NA, owner: type[object] | None = None
    ) -> None:
        self.value = value
        self.owner: type | None = owner
        self.name: str | None = repr(self)

    @override
    def __str__(self) -> str:
        res = []
        if self.owner is not None:
            res.append(f"{self.owner.__name__}")
        if self.name is not None:
            res.append(self.name)
        else:
            res.append("(unbound)")
        rep = ".".join(res)
        if self.value is not _NA:
            rep += "=" + repr(rep)
        return rep

    @override
    def __repr__(self) -> str:
        return f"<private property {self!s}>"

    def __get__(self, obj: object, cls: type[object]):
        if self.value in _NA:
            raise _NA.Error()
        return self.value

    def __set_name__(self, owner: type[object], name: str) -> None:
        self.name = name

    def __set__(self, obj: object, value: Any, /) -> None:
        if self.owner is not None:
            msg = f"{self.name} property is read-only"
            raise AttributeError(msg)
        self.value = value

    # __class_getitem__ = classmethod(types.GenericAlias)


class blocked:
    def __init__(self, fn: Callable[..., Any] | None = None) -> None:
        self.name = fn.__name__ if fn is not None else "method"
        self.owner = None

    def __set_name__(self, owner: type[object], name: str) -> None:
        self.name = name
        self.owner = owner

    def __get__(self, obj: object, cls: type[object]) -> Callable[..., Any]:
        if obj is None:
            return MethodType(self, cls)
        return MethodType(self, obj)

    def __call__(self, obj, *args, **kwargs):
        msg = f"{self.name} is blocked on {obj}"
        raise AttributeError(msg)



r"""Feature scaling layers."""

import typing

import torch
from torch import Tensor, nn


def layer_scale(
    input: Tensor, gamma: Tensor, training: bool, *, inplace: bool | None = None
) -> Tensor:
    """Functional version of :class:`LayerScale`."""
    if inplace is True or (
        inplace is None
        and (
            not torch.jit.is_scripting()
            and not torch.jit.is_tracing()
            and not torch.compiler.is_compiling()
            and not training
            and input.is_contiguous()
        )
    ):
        return input.mul_(gamma)
    return input * gamma


class LayerScale(nn.Module):
    """LayerScale methodology used in ConvNextV2."""

    def __init__(
        self,
        dim: int,
        gamma: float | Tensor = 1e-5,
        inplace: bool | None = False,
    ) -> None:
        r"""
        Parameters
        ----------
        dim : int
            The number of features in the input tensor.
        gamma : float or Tensor
            The initial value of the scaling factor.
        inplace : bool, optional
            Whether to perform the operation in-place or not. If `None`, the operation
            is performed in-place if the input tensor is contiguous and the
            module is *not* in training mode.
        """
        super().__init__()

        self.inplace = inplace
        self.gamma = nn.Parameter(gamma * torch.ones(dim))

        self.register_load_state_dict_pre_hook(self.__class__._patch_missing_gamma)

    def _patch_missing_gamma(
        self,
        state_dict: dict[str, Tensor],
        prefix: str,
        _local_metadata: dict[str, typing.Any],
        _strict: bool,
        _missing_keys: list[str],
        _unexpected_keys: list[str],
        _error_msgs: list[str],
    ) -> None:
        key_gamma = f"{prefix}gamma"
        if key_gamma not in state_dict:
            state_dict[key_gamma] = torch.ones_like(self.gamma)

    @typing.override
    def forward(self, input: Tensor) -> Tensor:
        return layer_scale(input, self.gamma, self.training, inplace=self.inplace)


class LayerScale2d(LayerScale):
    """Variant of :class:`LayerScale` for 2D inputs with shape ``(B, C, H, W)``."""

    @typing.override
    def forward(self, input: Tensor) -> Tensor:
        return layer_scale(
            input,
            self.gamma.view(1, -1, 1, 1),
            self.training,
            inplace=self.inplace,
        )



"""
Profiling utilities.
"""

import contextlib
import time
import types
import typing as T
from typing import override

import pandas as pd
import torch
from tensordict import TensorDict

from unipercept.types import Device

__all__ = [
    "LatencyAccumulator",
    "LatencyHandle",
    "LatencyProfile",
    "LatencyRecorder",
    "ProfileAccumulator",
    "latency_iter",
    "latency_profile",
    "profile",
    "profile_iter",
]

type Nanos = int
type LatencyMemory = T.MutableMapping[str, Nanos]

LATENCY_KEY_SEPARATOR: T.Final[str] = "."
# Backwards-compatible aliases for external callers
TimingsMemory = LatencyMemory
PROFILE_KEY_SEPARATOR = LATENCY_KEY_SEPARATOR


class LatencyProfile:
    """
    Context manager that profiles a block of code and stores the elapsed time.

    If the `dest` is `None` and the strict flag is set to `True`, an error is raised.
    Otherwise, the the null context manager is returned.
    """

    __slots__ = ("_start", "_time", "key", "memory")

    memory: LatencyMemory | None
    key: str
    _time: Nanos
    _start: Nanos | None

    @T.overload
    def __new__(
        cls,
        dest: None,
        key: str,
        /,
        *,
        strict: T.Literal[False],
    ) -> contextlib.AbstractContextManager[None]: ...

    @T.overload
    def __new__(
        cls,
        dest: None,
        key: str,
        /,
        *,
        strict: T.Literal[True],
    ) -> T.NoReturn: ...

    @T.overload
    def __new__(
        cls,
        dest: LatencyMemory,
        key: str,
        /,
        *,
        strict: bool = False,
    ) -> T.Self: ...

    def __new__(
        cls,
        dest: T.MutableMapping[str, Nanos] | None,
        key: str,
        /,
        *,
        strict: bool = False,
    ) -> contextlib.AbstractContextManager[None] | T.Self:
        """Return a profiling context that records nanosecond latency if enabled."""
        if dest is None:
            if strict:
                msg = f"No memory was passed and strict mode is enabled ({strict=})."
                raise RuntimeError(msg)
            return contextlib.nullcontext()

        # Construct the class
        self = super().__new__(cls)
        self.memory = dest
        self.key = key
        self._start = None

        return self

    @property
    def _running(self) -> bool:
        return self._start is not None

    @property
    def _delta(self) -> Nanos:
        if self._start is None:
            return 0
        return time.perf_counter_ns() - self._start

    def __enter__(self) -> "LatencyHandle":
        """Start timing and return a handle for nested measurements."""
        if self.memory is not None:
            self._start = time.perf_counter_ns()
        return LatencyHandle(self)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        """Stop timing and persist the measured latency."""
        if self.memory is None:
            return
        if not self._running:
            return
        self.memory[self.key] = self._delta
        self._start = None


class LatencyHandle(T.MutableMapping[str, Nanos]):
    """Redirect all operations to the parent mapping, adding a prefixed key."""

    __slots__ = ("_ref",)

    def __init__(self, ref: LatencyProfile) -> None:
        self._ref = ref

    @property
    def _prefix(self) -> str:
        return self._ref.key + LATENCY_KEY_SEPARATOR

    @property
    def memory(self) -> LatencyMemory:
        """Return the backing latency memory, raising if unavailable."""
        mem = self._ref.memory
        if mem is None:
            msg = "Leaf is not attached to a profile context with memory."
            raise ValueError(msg)
        return mem

    def cancel(self) -> None:
        """Cancel the ongoing profiling context."""
        self._ref._start = None

    @override
    def __setitem__(self, key: str, value: Nanos) -> None:
        self.memory[self._prefix + key] = value

    @override
    def __getitem__(self, key: str) -> Nanos:
        return self.memory[self._prefix + key]

    @override
    def __delitem__(self, key: str) -> None:
        del self.memory[self._prefix + key]

    @override
    def __iter__(self) -> T.Iterator[str]:
        return iter(self.memory)

    @override
    def __len__(self) -> int:
        return len(self.memory)

    @override
    def keys(self) -> T.KeysView[str]:
        return self.memory.keys()

    @override
    def values(self) -> T.ValuesView[int]:
        return self.memory.values()

    @override
    def items(self) -> T.ItemsView[str, int]:
        return self.memory.items()


@T.overload
def latency_iter[R: T.Any](
    dest: LatencyMemory | None,
    key: str,
    src: T.Iterator[R] | T.Iterable[R],
    *,
    warmup: int,
    strict: T.Literal[False],
) -> T.Iterable[R]: ...


@T.overload
def latency_iter[R: T.Any](
    dest: None,
    key: str,
    src: T.Iterator[R] | T.Iterable[R],
    *,
    warmup: int,
    strict: T.Literal[True],
) -> T.NoReturn: ...


@T.overload
def latency_iter[R: T.Any](
    dest: LatencyMemory,
    key: str,
    src: T.Iterator[R] | T.Iterable[R],
    *,
    warmup: int,
    strict: bool = False,
) -> T.Iterable[R]: ...


def latency_iter[R: T.Any](
    dest: LatencyMemory | None,
    key: str,
    src: T.Iterator[R] | T.Iterable[R],
    *,
    warmup: int = 0,
    strict: bool = False,
) -> T.Iterable[R]:
    r"""Profile item retrieval latency for an iterable.

    Records nanosecond latency per item and yields the items unchanged.

    Parameters
    ----------
    timings : TimingsMemory | None
        The memory to store the timings in.
    source : Iterator[_R] | Iterble[_R]
        The source to iterate over. If it is not already an iterator, we turn it into
        one using `iter(source)`.
    key : str
        The key to store the timings under in the target memory.
    warmup : int
        The number of iterations to skip before recording the timings.
    """
    assert warmup >= 0, f"Warmup ({warmup=}) must be a non-negative integer."
    if dest is None:
        if strict:
            msg = f"No memory was passed and strict mode is enabled ({strict=})."
            raise RuntimeError(msg)
        yield from src
        return

    iterator = src if isinstance(src, T.Iterator) else iter(src)
    remaining_warmup = warmup

    while True:
        with LatencyProfile(dest, key, strict=True) as ph:
            try:
                item = next(iterator)
            except StopIteration:
                ph.cancel()
                break

            if remaining_warmup > 0:
                ph.cancel()
                remaining_warmup -= 1
        yield item


class ProfileRecord(T.NamedTuple):
    r"""A single long-format record used to store profiling data."""

    key: str
    time: int | float


class LatencyRecorder(T.MutableMapping[str, int]):
    """Mapping that records latencies and offers helper entrypoints."""

    __slots__ = ("records",)

    _KEY_TIME: T.Final[str] = "seconds"
    _KEY_NAME: T.Final[str] = "name"

    records: list[ProfileRecord]

    def __init__(self) -> None:
        self.records = []

    def profile(self, key: str, *, strict: bool = False) -> LatencyProfile:
        """Return a latency profiling context bound to this recorder."""
        return LatencyProfile(self, key, strict=strict)

    def iter(
        self,
        key: str,
        src: T.Iterator[T.Any] | T.Iterable[T.Any],
        *,
        warmup: int = 0,
        strict: bool = False,
    ) -> T.Iterable[T.Any]:
        """Profile item retrieval latency for an iterable, bound to this recorder."""
        return latency_iter(self, key, src, warmup=warmup, strict=strict)

    def run(
        self, warmup: int = 0, incomplete_ok: bool = True, enabled: bool = True
    ) -> T.Iterator[dict[str, int] | None]:
        """
        Return an iterator of dicts to which profiling data can be written.

        The first populated dict defines the expected keys for subsequent iterations.
        """
        if not enabled:
            while True:
                yield None
        skip = warmup
        while True:
            target = {}
            yield target
            if len(target.keys()) == 0:
                continue

            new_keys = frozenset(target.keys())
            cur_keys = frozenset(self.keys())
            if len(self.records) > 0 and new_keys != cur_keys and not incomplete_ok:
                missing = new_keys - cur_keys
                msg = (
                    f"Received {missing} keys that were not present in the previous "
                    "iteration."
                )
                raise ValueError(msg)
            if skip > 0:
                skip -= 1
            else:
                for k, v in target.items():
                    self[k] = v
            del target

    @override
    def __getitem__(self, key: str) -> float:
        return self.to_means()[key]

    @override
    def __setitem__(self, key: str, value: int) -> None:
        assert isinstance(value, int), "Value must be time in nanoseconds"
        rec = ProfileRecord(key, value)
        self.records.append(rec)

    @override
    def __delitem__(self, key: str) -> None:
        idxs_to_remove = [i for i, r in enumerate(self.records) if r.key == key]
        for i in reversed(idxs_to_remove):
            del self.records[i]

    @override
    def __iter__(self) -> T.Iterator[str]:
        return iter(self.to_means())

    @override
    def __len__(self) -> int:
        return len(self.records)

    def reset(self) -> None:
        """Clear all recorded latencies."""
        self.records.clear()

    def to_records(self, *, unit: bool = True) -> T.Iterable[ProfileRecord]:
        """Return records in either seconds (default) or raw nanoseconds."""
        return (
            ProfileRecord(r.key, r.time * 1e-9 if unit else r.time)
            for r in self.records
        )

    def to_dataframe(self, *, unit: bool = True) -> pd.DataFrame:
        """Return the profiling data as a pandas DataFrame."""
        return pd.DataFrame.from_records(
            self.to_records(unit=unit),
            columns=[self._KEY_NAME, self._KEY_TIME],
        )

    def to_tensordict(
        self, *, device: Device | None = None, unit: bool = True
    ) -> TensorDict:
        """Return the profiling data grouped in a TensorDict."""
        return TensorDict(
            {
                key: torch.tensor(
                    times,
                    device=device,
                    dtype=torch.double if unit else torch.long,
                    requires_grad=False,
                )
                for key, times in self.to_groups(unit=unit).items()
            },
            device=device,
            batch_size=[],
        )

    def to_summary(self, *, unit: bool = True) -> pd.DataFrame:
        """Return summarizing statistics for every key in the accumulator."""
        df = self.to_dataframe(unit=unit)

        return df.groupby(self._KEY_NAME)[self._KEY_TIME].agg(
            ["count", "sum", "mean", "std", "min", "max", "median"]
        )

    def to_means(self) -> dict[str, float]:
        """Return mean values per recorded key."""
        return {k: sum(v) / len(v) for k, v in self.to_groups().items()}

    def to_groups(self, *, unit: bool = True) -> dict[str, list[int | float]]:
        """Group the records by their key."""
        groups = {}
        for r in self.records:
            t = r.time * 1e-9 if unit else r.time
            groups.setdefault(r.key, []).append(t)
        return groups

    @override
    def items(self) -> T.ItemsView[str, float]:
        return self.to_means().items()

    @override
    def keys(self) -> T.AbstractSet[str]:
        return frozenset(r.key for r in self.records)

    @override
    def values(self) -> T.ValuesView[float]:
        return self.to_means().values()

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(self.keys())})"



import enum as E
import typing as T

import torch
import torch.fx
import typing_extensions as TX
from torch import Tensor, nn


class DropPathMode(E.StrEnum):
    """The drop mode for :func:`drop_path`."""

    BATCH = "batch"
    ROW = "row"


def drop_path(
    input: Tensor,
    p: float,
    training: bool,
    *,
    mode: DropPathMode | str = DropPathMode.ROW,
) -> Tensor:
    """
    Implements stochastic depth from [1].

    Parameters
    ----------
    input : Tensor[*shape]
        The input tensor.
    p : float
        The drop probability.
    mode : DropPathMode
        The drop mode, either "batch" or "row".
        In "batch" mode, the same mask is applied to all elements in the batch.
        In "row" mode, a different mask is applied to each row in the batch.
    training : bool, optional
        Whether the model is in training mode, by default True.

    Returns
    -------
    Tensor[*shape]
        The output tensor.

    References
    ----------
    [1] `"Deep Networks with Stochastic Depth" <https://arxiv.org/abs/1603.09382>`_.
    """
    if not training or p == 0.0:
        return input

    match mode:
        case DropPathMode.BATCH:
            size = [1] * input.ndim
        case DropPathMode.ROW:
            size = [input.shape[0]] + [1] * (input.ndim - 1)
        case _:
            msg = f"Drop path mode not implemented: {mode}"
            raise NotImplementedError(msg)

    survival_rate = 1.0 - p
    noise = input.new_empty(size).bernoulli_(survival_rate)
    if survival_rate > 0.0:
        noise.div_(survival_rate)
    return input * noise


torch.fx.wrap("drop_path")


class DropPath(nn.Module):
    """Module that wraps :func:`drop_path`."""

    def __init__(
        self,
        p: float,
        mode: DropPathMode | T.Literal["row", "batch"] = DropPathMode.ROW,
    ) -> None:
        """
        Parameters
        ----------
        p : float
            The drop probability.
        mode : DropPathMode, default DropPathMode.ROW
            Drop mode. See :func:`drop_path` for details.
        """
        super().__init__()
        self.p = p
        self.mode = DropPathMode(mode)

    @TX.override
    def forward(self, input: Tensor) -> Tensor:
        """
        Applies drop path to the input tensor.

        Parameters
        ----------
        input : Tensor[*shape]
            The input tensor.

        Returns
        -------
        Tensor[*shape]
            The output tensor.
        """
        return drop_path(input, self.p, self.training, mode=self.mode)

    @TX.override
    def extra_repr(self) -> str:
        return f"p={self.p}, mode={self.mode}"



from __future__ import annotations

from typing import override

import torch
from torch import Tensor, nn

from . import render


class DepthProjection(nn.Module):
    """Maps the world-positon of an object using a projection of the object 2D mask (i.e. segmentation) and estimated mean  depth."""

    max_depth: Tensor

    def __init__(
        self,
        max_depth: Tensor | float,
        scale: float = 4.0,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        max_depth
            Maximum depth for denoramlization of the mean depth, which should be
            fixed for a given dataset.
        scale
            Value to upscale screen by, e.g. when the network gives masks
            with size (h,w) then they will be projected as if they had size
            (h*scale, w*scale). This is useful because predictions are often
            made on a downscaled version of the image.
        """
        super().__init__(**kwargs)

        self.scale = scale
        self.threshold = 0.2

        self.register_buffer(
            "max_depth",
            torch.tensor(max_depth, dtype=torch.float32, requires_grad=False),
            persistent=False,
        )

    @override
    def forward(
        self, logits: Tensor, mean_depths: Tensor, cams: render.Cameras
    ) -> Tensor:
        """ "
        Project the masks and mean depths to world coordinates.

        Parameters
        ----------
        masks
            Instance masks of shape (n, h, w).
        mean_depths
            Mean depth of each instance of shape (n,).
        K
            Camera intrinsics of shape (3, 3).

        Returns
        -------
            Points in world coordinates of shape (n, 3).
        """
        masks = logits.sigmoid() > self.threshold
        if len(masks) == 0:
            return torch.empty((0, 3), dtype=self.max_depth.dtype, device=masks.device)

        points_xyd = self._read_points_xyd(masks, mean_depths)
        return self._unproject(cams, points_xyd)

    def _read_points_xyd(self, masks: Tensor, mean_depths: Tensor) -> Tensor:
        """
        Create a list of points using the mass center and mean depth of each
        instance.
        """
        assert masks.ndim == 3, masks.ndim  # n, h, w
        assert mean_depths.ndim == 2, mean_depths.ndim  # n x d
        assert len(masks) == len(mean_depths), (len(masks), len(mean_depths))

        indices, yx = masks.argwhere().split([1, 2], dim=1)
        _, counts = torch.unique(indices, return_counts=True)
        indices_len: list[int] = counts.tolist()

        # Convert yx -> xy
        xy = torch.flip(yx, [1])

        # Compensate for downscaling of masks
        xy = xy * self.scale

        # Split indices
        indices_split = indices.float().split(indices_len)
        for i in indices_split:
            assert i.std() == 0

        xy_split = xy.float().split(indices_len)

        # Compute means (mass center)
        xy_means = [torch.mean(xy, dim=0) for xy in xy_split]

        assert len(xy_means) == len(mean_depths), (
            len(xy_means),
            len(mean_depths),
        )

        return torch.cat(
            [torch.stack(xy_means), mean_depths],
            dim=1,
        )

    def _unproject(self, cams: render.Cameras, points_xyd: Tensor) -> Tensor:
        points_xyd_batch = points_xyd[None, :]
        points_world = cams.unproject_points(points_xyd_batch)
        return points_world[0]



"""Modules for working with sequentual layers."""

import typing as T

import torch
import torch.nn


class SequentialList(torch.nn.Sequential):
    """A nn.Sequential that takes a list of tensors as input and returns a list of tensors as output."""

    @T.override
    def forward(self, input: list[torch.Tensor]) -> list[torch.Tensor]:
        for module in self:
            input = module(input)
        return input



"""This module hosts various losses for perception tasks."""

from unipercept.utils.python import defer_imports as _defer_imports

__all__ = [
    "_stable",
    "contrastive",
    "depth",
    "focal",
    "functional",
    "guided",
    "image",
    "mixins",
    "panoptic",
]
__getattr__, __dir__ = _defer_imports(__name__, __all__)



"""Extends ``torchvision.tv_tensors`` with new datapoint types."""

from . import helpers, registry
from ._camera import *
from ._depth import *
from ._flow import *
from ._segmentation import *
from ._torchvision import *



from typing import Final, override

import torch
from torch import nn

from unipercept.types import Tensor


class ReconstructionLoss(nn.Module):
    alpha: Final[float]

    def __init__(self, alpha=0.85) -> None:
        super().__init__()
        self.alpha = alpha
        self.ssim = SSIMLoss()
        self.l1 = nn.L1Loss(reduction="none")

    @override
    def forward(self, output: torch.Tensor, target: torch.Tensor):
        ssim = self.ssim(output, target).mean(
            dim=1, keepdim=True
        )  # (N, 1, H, W), average channels
        l1 = self.l1(output, target).mean(
            dim=1, keepdim=True
        )  # (N, 1, H, W), average channels
        return (
            0.5 * self.alpha * ssim + (1 - self.alpha) * l1
        )  # NOTE: ssim is already (1-ssim)


class SSIMLoss(nn.Module):
    """
    Calculate the Structural Similarity Error between two vectors.

    .. math::

        SSIM(x,y) = [l(x,y)]^a * [c(x,y)]^b * [s(x,y)]^y

    where l:luminance, c:contrast, s:structure

    According to [1], we set a=b=y=1, and then

    .. math::

        SSIM(x,y) = (2*m_x*m_y+c1)(2*s_xy+c2)/(m_x^2 + m_y^2 + c1)(s_x_2 _s_y^2 + c2)

    References
    ----------
    [1] Wang, Zhou, Eero P. Simoncelli, and Alan C. Bovik. "Multiscale structural
        similarity for image quality assessment."
        Signals, Systems and Computers, 2004.
    """

    c1: Final[float]
    c2: Final[float]

    def __init__(self) -> None:
        super().__init__()

        bits_per_pixel = 8  # pixel values range between 0 and 255
        L = 2**bits_per_pixel - 1  # dynamic range of the pixel values
        k1 = 0.01  # by default
        k2 = 0.03  # by default
        self.c1 = (k1 * L) ** 2  # stabilize the division with weak denominator
        self.c2 = (k2 * L) ** 2  # purpose same as c1

        self.avg_pool = nn.AvgPool2d(kernel_size=3, stride=1)
        self.refl_pad = nn.ReflectionPad2d(1)

    @override
    def forward(self, src: Tensor, tgt: Tensor):
        # Add 1 reflection padding in all boundaries of each output, and target.
        # NOTE: Padding because AvgPool2d will decrease the dim = (N, B, H, W) to dim = (N, B, H-1, W-1).
        # NOTE: Reflection because we use AvgPool2d to calculate the mean and (cov)variance on 3x3 blocks.
        src = self.refl_pad(src)
        tgt = self.refl_pad(tgt)

        # Calculate mean and variance for output.
        output_mu = self.avg_pool(src)  # pixel sample mean of output
        output_sigma = self.avg_pool(src**2) - output_mu  # variance of output

        # Calculate mean and variance for target.
        target_mu = self.avg_pool(tgt)  # pixel sample mean of target
        target_sigma = self.avg_pool(tgt**2) - target_mu  # variance of target

        # Calculate covvariance between output and target.
        output_target_sigma = (
            self.avg_pool(src * tgt) - output_mu * target_mu
        )  # covariance of output and target

        # Numerator of SSIM
        numerator = (2 * output_mu * target_mu + self.c1) * (
            2 * output_target_sigma + self.c2
        )
        # Denominator of SSIM
        denominator = (output_mu**2 + target_mu**2 + self.c1) * (
            output_sigma + target_sigma + self.c2
        )
        # Compute SSIM
        _ssim = numerator / denominator  # ssim \in [-1, 1]

        # Normalize SSIM
        norm_ssim = (1 - _ssim) / 2  # ssim \in [0, 1]

        # Clamp to avoid potential overflow.
        return torch.clamp(norm_ssim, min=0, max=1)



r"""
Depth tensors.
=============

Provides support for depth tensors, e.g. disparity maps, depth maps, etc.
"""

import enum as E
import functools
import io
import json
import typing as T

import expath
import numpy as np
import PIL.Image as pil_image
import safetensors.torch as safetensors
import torch
import torchvision.io
from einops import rearrange
from tensordict import MemoryMappedTensor
from torch.nn.functional import interpolate
from torch.types import Device
from torchvision.transforms.v2.functional import register_kernel
from torchvision.transforms.v2.functional._geometry import _compute_resized_output_size
from torchvision.tv_tensors import Mask

from unipercept.tensors.helpers import get_kwd, read_pixels, write_png_l16
from unipercept.tensors.registry import pixel_maps
from unipercept.types import Device, DType, Pathable, Tensor
from unipercept.utils.inspect import locate_object

__all__ = [
    "DepthFormat",
    "DepthMode",
    "DepthTensor",
    "downsample_depthmap",
    "load_depthmap",
    "resize_depthmap",
    "save_depthmap",
]

DEFAULT_DEPTH_DTYPE: T.Final = torch.float32


class DepthFormat(E.StrEnum):
    r"""Enum class for depth map file formats and their respective mode."""

    TIFF = E.auto()
    DEPTH_INT16 = E.auto()
    DISPARITY_INT16 = E.auto()
    TORCH = E.auto()
    SAFETENSORS = E.auto()
    MEMMAP = E.auto()


class DepthMode(E.StrEnum):
    r"""Enum class for depth prediction modes."""

    ABSOLUTE = E.auto()
    DISPARITY = E.auto()


@pixel_maps.register()
class DepthTensor(Mask):
    r"""Represents a two-dimensional depth map as a PyTorch tensor."""

    def __new__(
        cls,
        data: T.Any,
        *,
        dtype: DType | None = None,
        device: Device = None,
        requires_grad: bool | None = None,
    ) -> T.Self:
        tensor = cls._to_tensor(
            data, dtype=dtype, device=device, requires_grad=requires_grad
        )
        return tensor.as_subclass(cls)

    @classmethod
    def default_like(cls, other: Tensor) -> T.Self:
        """Returns a default instance of this class with the same shape as the given tensor."""
        return cls(torch.full_like(other, fill_value=0, dtype=torch.float32))

    @classmethod
    def default(cls, shape: torch.Size, device: Device = "cpu") -> T.Self:
        """Returns a default instance of this class with the given shape."""
        return cls(torch.zeros(shape, device=device, dtype=torch.float32))  # type: ignore


@torch.no_grad()
def save_depthmap(
    data: Tensor, path: Pathable, format: DepthFormat | str | None = None
) -> None:
    import expath

    data = data.detach().cpu()

    path = expath.locate(path)
    if format is None:
        match path.suffix.lower():
            case ".tiff":
                format = DepthFormat.TIFF
            case ".pth", ".pt":
                format = DepthFormat.TORCH
            case ".safetensors":
                format = DepthFormat.SAFETENSORS
            case ".memmap":
                format = DepthFormat.MEMMAP
            case _:
                msg = f"Could not infer depth format from path: {path}"
                raise ValueError(msg)

    path.parent.mkdir(parents=True, exist_ok=True)

    match DepthFormat(format):
        case DepthFormat.TIFF:
            depth_image = pil_image.fromarray(data.float().squeeze_(0).cpu().numpy())
            if depth_image.mode != "F":
                msg = f"Expected image format 'F'; Got {depth_image.mode!r}"
                raise ValueError(msg)
            assert path.suffix.lower() == ".tiff", path
            depth_image.save(path, format="TIFF")
        case DepthFormat.SAFETENSORS:
            assert path.suffix.lower() == ".safetensors", path
            safetensors.save_file({"data": torch.as_tensor(data)}, path)
        case DepthFormat.TORCH:
            torch.save(torch.as_tensor(data), path)
        case DepthFormat.DEPTH_INT16:
            # depth_image = (self * float(2**8)).numpy().astype(np.uint16)
            # image = pil_image.fromarray(depth_image, mode="I;16")
            # image.save(path)
            assert path.suffix.lower() == ".png", path
            write_png_l16(path, data * float(2**8))
        case _:
            msg = f"Unsupported depth format: {format}"
            raise NotImplementedError(msg)


@torch.no_grad()
def load_depthmap(
    input: expath.PathType | Tensor,
    dtype: torch.dtype = DEFAULT_DEPTH_DTYPE,
    **meta_kwds: T.Any,
) -> DepthTensor:
    if not isinstance(input, torch.Tensor):
        input = str(expath.locate(input))
    # Switch by depth format
    format = get_kwd(meta_kwds, "format", DepthFormat | str)
    match DepthFormat(format):  # type: ignore
        case DepthFormat.TIFF:
            if isinstance(input, torch.Tensor):
                fp_open = functools.partial(io.BytesIO, input.numpy().tobytes())
            else:
                fp_open = functools.partial(expath.open, input, "rb")
            with fp_open() as fp, pil_image.open(fp) as img:
                assert img.mode == "F", img.mode
                dm_np = np.array(img, copy=True)
            dm = torch.from_numpy(dm_np)
            assert dm.ndim == 2, dm.shape
            assert dm.dtype == torch.float32, dm.dtype
            dm = dm.unsqueeze(0)
        case DepthFormat.DEPTH_INT16:
            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            dm = torchvision.io.read_image(
                input, mode=torchvision.io.ImageReadMode.UNCHANGED
            )
            assert dm.ndim == 3, dm.shape
            assert dm.dtype == torch.uint16, dm.dtype
            dm = dm.to(dtype=dtype) / float(2**8)
        case DepthFormat.DISPARITY_INT16:
            dm = load_depthmap_from_disparitymap(input, **meta_kwds)
        case DepthFormat.SAFETENSORS:
            if isinstance(input, torch.Tensor):
                msg = "Expected a path to a file, not a tensor."
                raise TypeError(msg)
            input = str(expath.locate(input))
            dm = safetensors.load_file(input)["data"]
        case DepthFormat.TORCH:
            if isinstance(input, torch.Tensor):
                msg = "Expected a path to a file, not a tensor."
                raise TypeError(msg)
            input = str(expath.locate(input))
            dm = torch.load(input, map_location="cpu")
        case DepthFormat.MEMMAP:
            if isinstance(input, torch.Tensor):
                msg = "Expected a path to a file, not a tensor."
                raise TypeError(msg)
            input = expath.locate(input)
            tensor_name = input.stem
            path_meta = input.parent / "meta.json"
            with path_meta.open("r") as f:
                meta = json.load(f)[tensor_name]

            dm = MemoryMappedTensor.from_filename(
                filename=input,
                dtype=locate_object(meta["dtype"]),
                shape=torch.Size(meta["shape"]),
            ).contiguous()
        case _:
            msg = f"Unsupported depth format: {format}"
            raise NotImplementedError(msg)

    # TODO: Add angular FOV compensation via metadata
    dm = dm.to(dtype=dtype)
    dm[dm == torch.inf] = 0.0
    dm[dm == torch.nan] = 0.0
    if dm.ndim != 3:
        msg = (
            f"Depth map has {dm.ndim} dimensions, expected  3 (CHW or HWC format). "
            f"Got {dm.shape} instead."
        )
        raise ValueError(msg)

    return dm.as_subclass(DepthTensor)


def load_depthmap_from_disparitymap(
    path: str,
    camera_baseline: float,
    camera_fx: float,
) -> DepthTensor:
    # Get machine epsilon for the given dtype, used to check for invalid values
    eps = torch.finfo(torch.float32).eps

    # Read disparity map
    disp = read_pixels(path, False)
    assert disp.dtype == torch.int32, disp.dtype

    # Convert disparity from 16-bit to float
    disp = disp.to(dtype=torch.float32, copy=False)
    disp -= 1
    disp[disp >= eps] /= 256.0

    # Infer depth using camera parameters and disparity
    valid_mask = disp >= eps

    depth = torch.zeros_like(disp)
    depth[valid_mask] = (camera_baseline * camera_fx) / disp[valid_mask]

    # Set invalid depth values to 0
    depth[depth == torch.inf] = 0
    depth[depth == torch.nan] = 0

    return T.cast(DepthTensor, depth.as_subclass(DepthTensor))


class DepthDownsampleMethod(E.StrEnum):
    MEDIAN = E.auto()
    NEAREST = E.auto()


def downsample_depthmap(
    depth_map: Tensor,
    size: tuple[int, int] | torch.Size,
    method: DepthDownsampleMethod | str = DepthDownsampleMethod.MEDIAN,
) -> Tensor:
    """
    Downsampling of depth maps.

    Parameters
    ----------
    depth_map: Tensor[..., H_old, W_old]
        The input depth map.
    size : tuple[H_new, W_new]
        The target size of the downsampled depth map.
    method : DepthDownsampleMethod or str
        The method used for downsampling. Default: ```DepthDownsampleMethod.MEDIAN``.

    Returns
    -------
    Tensor[..., H, W]
        The downsampled depth map.
    """
    # Check that the shape is actually divisible by the target size, else do NN downsample first to the closest size
    h_old, w_old = depth_map.shape[-2:]
    h_new, w_new = size

    match DepthDownsampleMethod(method):
        case DepthDownsampleMethod.MEDIAN:
            if h_old % h_new != 0 or w_old % w_new != 0:
                h_new = h_old // round(h_old / h_new)
                w_new = w_old // round(w_old / w_new)
                depth_map = interpolate_depth(depth_map, (h_new, w_new))

            # Perform median pooling
            depth_map = rearrange(
                depth_map,
                "... (h1 h2) (w1 w2) -> ... h1 w1 (h2 w2)",
                h1=h_new,
                w1=w_new,
            )
            depth_map[depth_map <= 0] = torch.nan
            depth_map = torch.nanmedian(depth_map, dim=-1).values

            # Set invalid depth values to 0
            depth_map[~torch.isfinite(depth_map)] = 0
        case DepthDownsampleMethod.NEAREST:
            depth_map = interpolate_depth(depth_map, size)

    return depth_map


@register_kernel(functional="resize", tv_tensor_cls=DepthTensor)
def resize_depthmap(
    image: DepthTensor,
    size: list[int],
    interpolation: T.Any = None,
    max_size: int | None = None,
    antialias: T.Any = False,
    use_rescale: bool = False,
) -> Tensor:
    shape = image.shape
    h_old, w_old = shape[-2:]
    h_new, w_new = _compute_resized_output_size(
        (h_old, w_old), size=size, max_size=max_size
    )

    if h_new <= h_old and w_new <= w_old:
        res = downsample_depthmap(image, (h_new, w_new))
    else:
        res = interpolate_depth(image, (h_new, w_new))

    if use_rescale:
        d_min = image.min()
        d_max = image.max()
        scale = (h_old / h_new + w_old / w_new) / 2

        res = (res * scale).clamp(d_min, d_max)

    return res.as_subclass(DepthTensor)


def interpolate_depth(
    depth_map: Tensor,
    size: tuple[int, int] | torch.Size,
) -> Tensor:
    """
    Quick wrapper for 2D nearest neighbor interpolation, if the input does not have enough
    dimensions, then these are added before interpolation and removed at the end.
    """
    ndim = depth_map.ndim
    while depth_map.ndim < 4:
        depth_map = depth_map.unsqueeze(0)
    depth_map = interpolate(depth_map, size=size, mode="nearest-exact")
    while depth_map.ndim > ndim:
        depth_map = depth_map.squeeze(0)

    return depth_map



"""Wrap and register torchvision's tensor types."""

import enum as E
import typing as T

import expath
import safetensors
import torch
import torchvision.io
from torchvision.tv_tensors import (
    BoundingBoxes,
    BoundingBoxFormat,
)
from torchvision.tv_tensors import (
    Image as ImageTensor,
)
from torchvision.tv_tensors import (
    Mask as MaskTensor,
)

from unipercept.types import Tensor

from .registry import pixel_maps

__all__ = [
    "BoundingBoxFormat",
    "BoundingBoxes",
    "ImageTensor",
    "MaskFormat",
    "MaskMeta",
    "MaskTensor",
    "load_image",
    "load_mask",
    "save_mask",
]

########
# Mask #
########

pixel_maps.add(MaskTensor)


class MaskFormat(E.StrEnum):
    PNG_L = E.auto()
    PNG_LA = E.auto()
    PNG_L16 = E.auto()
    PNG_LA16 = E.auto()
    TORCH = E.auto()
    SAFETENSORS = E.auto()


class MaskMeta(T.TypedDict):
    format: T.NotRequired[MaskFormat]


def load_mask(
    input: expath.PathType | torch.Tensor, /, **kwargs: MaskMeta
) -> MaskTensor:
    format = kwargs["format"]

    match format:
        case MaskFormat.PNG_L | MaskFormat.PNG_L16:
            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            return torchvision.io.decode_png(
                input, mode=torchvision.io.ImageReadMode.GRAY
            ).as_subclass(MaskTensor)
        case MaskFormat.PNG_LA | MaskFormat.PNG_LA16:
            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            return torchvision.io.decode_png(
                input, mode=torchvision.io.ImageReadMode.GRAY
            ).as_subclass(MaskTensor)
        case MaskFormat.TORCH:
            if isinstance(input, torch.Tensor):
                msg = "Expected a path to a file, not a tensor."
                raise TypeError(msg)
            with expath.open(input, "rb") as fh:
                return torch.load(fh).as_subclass(MaskTensor)
        case MaskFormat.SAFETENSORS:
            if isinstance(input, torch.Tensor):
                msg = "Expected a path to a file, not a tensor."
                raise TypeError(msg)
            input = str(expath.locate(input))
            return safetensors.torch.load(input).as_subclass(MaskTensor)
        case _:
            msg = f"Unsupported format: {format}"
            raise NotImplementedError(msg)


def save_mask(
    tensor: Tensor | MaskTensor, path: expath.PathType, /, **kwargs: MaskMeta
) -> MaskTensor:
    msg = "Not implemented yet."
    raise NotImplementedError(msg)


#########
# Image #
#########

pixel_maps.add(ImageTensor)


def load_image(input: expath.PathType | torch.Tensor) -> ImageTensor:
    """Reads an image from the given path."""
    from torchvision.transforms.v2.functional import to_dtype, to_image

    if not isinstance(input, torch.Tensor):
        input = str(expath.locate(input))
    img = torchvision.io.decode_image(input)
    img = to_image(img)
    img = to_dtype(img, torch.float32, scale=True)

    assert img.shape[0] == 3, f"Expected image to have 3 channels, got {img.shape[0]}!"

    return img.as_subclass(ImageTensor)



from __future__ import annotations

import warnings

import torch
from torch import Tensor

from ._acos import acos_linear_extrapolation


def so3_relative_angle(
    R1: Tensor,
    R2: Tensor,
    cos_angle: bool = False,
    cos_bound: float = 1e-4,
    eps: float = 1e-4,
) -> Tensor:
    """
    Calculates the relative angle (in radians) between pairs of
    rotation matrices `R1` and `R2` with `angle = acos(0.5 * (Trace(R1 R2^T)-1))`.

    .. note::
        This corresponds to a geodesic distance on the 3D manifold of rotation
        matrices.

    Args:
        R1: Batch of rotation matrices of shape `(minibatch, 3, 3)`.
        R2: Batch of rotation matrices of shape `(minibatch, 3, 3)`.
        cos_angle: If==True return cosine of the relative angle rather than
            the angle itself. This can avoid the unstable calculation of `acos`.
        cos_bound: Clamps the cosine of the relative rotation angle to
            [-1 + cos_bound, 1 - cos_bound] to avoid non-finite outputs/gradients
            of the `acos` call. Note that the non-finite outputs/gradients
            are returned when the angle is requested (i.e. `cos_angle==False`)
            and the rotation angle is close to 0 or π.
        eps: Tolerance for the valid trace check of the relative rotation matrix
            in `so3_rotation_angle`.

    Returns
    -------
        Corresponding rotation angles of shape `(minibatch,)`.
        If `cos_angle==True`, returns the cosine of the angles.

    Raises
    ------
        ValueError if `R1` or `R2` is of incorrect shape.
        ValueError if `R1` or `R2` has an unexpected trace.
    """
    R12 = torch.bmm(R1, R2.permute(0, 2, 1))
    return so3_rotation_angle(R12, cos_angle=cos_angle, cos_bound=cos_bound, eps=eps)


def so3_rotation_angle(
    R: Tensor,
    eps: float = 1e-4,
    cos_angle: bool = False,
    cos_bound: float = 1e-4,
) -> Tensor:
    """
    Calculates angles (in radians) of a batch of rotation matrices `R` with
    `angle = acos(0.5 * (Trace(R)-1))`. The trace of the
    input matrices is checked to be in the valid range `[-1-eps,3+eps]`.
    The `eps` argument is a small constant that allows for small errors
    caused by limited machine precision.

    Args:
        R: Batch of rotation matrices of shape `(minibatch, 3, 3)`.
        eps: Tolerance for the valid trace check.
        cos_angle: If==True return cosine of the rotation angles rather than
            the angle itself. This can avoid the unstable
            calculation of `acos`.
        cos_bound: Clamps the cosine of the rotation angle to
            [-1 + cos_bound, 1 - cos_bound] to avoid non-finite outputs/gradients
            of the `acos` call. Note that the non-finite outputs/gradients
            are returned when the angle is requested (i.e. `cos_angle==False`)
            and the rotation angle is close to 0 or π.

    Returns
    -------
        Corresponding rotation angles of shape `(minibatch,)`.
        If `cos_angle==True`, returns the cosine of the angles.

    Raises
    ------
        ValueError if `R` is of incorrect shape.
        ValueError if `R` has an unexpected trace.
    """
    _N, dim1, dim2 = R.shape
    if dim1 != 3 or dim2 != 3:
        msg = "Input has to be a batch of 3x3 Tensors."
        raise ValueError(msg)

    rot_trace = R[:, 0, 0] + R[:, 1, 1] + R[:, 2, 2]

    if ((rot_trace < -1.0 - eps) + (rot_trace > 3.0 + eps)).any():
        msg = "A matrix has trace outside valid range [-1-eps,3+eps]."
        raise ValueError(msg)

    # phi ... rotation angle
    phi_cos = (rot_trace - 1.0) * 0.5

    if cos_angle:
        return phi_cos
    if cos_bound > 0.0:
        bound = 1.0 - cos_bound
        return acos_linear_extrapolation(phi_cos, (-bound, bound))
    return torch.acos(phi_cos)


def so3_exp_map(log_rot: Tensor, eps: float = 0.0001) -> Tensor:
    """
    Convert a batch of logarithmic representations of rotation matrices `log_rot`
    to a batch of 3x3 rotation matrices using Rodrigues formula [1].

    In the logarithmic representation, each rotation matrix is represented as
    a 3-dimensional vector (`log_rot`) who's l2-norm and direction correspond
    to the magnitude of the rotation angle and the axis of rotation respectively.

    The conversion has a singularity around `log(R) = 0`
    which is handled by clamping controlled with the `eps` argument.

    Args:
        log_rot: Batch of vectors of shape `(minibatch, 3)`.
        eps: A float constant handling the conversion singularity.

    Returns
    -------
        Batch of rotation matrices of shape `(minibatch, 3, 3)`.

    Raises
    ------
        ValueError if `log_rot` is of incorrect shape.

    [1] https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    """
    return _so3_exp_map(log_rot, eps=eps)[0]


def so3_exponential_map(log_rot: Tensor, eps: float = 0.0001) -> Tensor:
    warnings.warn(
        """so3_exponential_map is deprecated,
        Use so3_exp_map instead.
        so3_exponential_map will be removed in future releases.""",
        PendingDeprecationWarning,
        stacklevel=2,
    )

    return so3_exp_map(log_rot, eps)


def _so3_exp_map(
    log_rot: Tensor, eps: float = 0.0001
) -> tuple[Tensor, Tensor, Tensor, Tensor]:
    """
    A helper function that computes the so3 exponential map and,
    apart from the rotation matrix, also returns intermediate variables
    that can be re-used in other functions.
    """
    _, dim = log_rot.shape
    if dim != 3:
        msg = "Input tensor shape has to be Nx3."
        raise ValueError(msg)

    nrms = (log_rot * log_rot).sum(1)
    # phis ... rotation angles
    rot_angles = torch.clamp(nrms, eps).sqrt()
    # pyre-fixme[58]: `/` is not supported for operand types `float` and `Tensor`.
    rot_angles_inv = 1.0 / rot_angles
    fac1 = rot_angles_inv * rot_angles.sin()
    fac2 = rot_angles_inv * rot_angles_inv * (1.0 - rot_angles.cos())
    skews = hat(log_rot)
    skews_square = torch.bmm(skews, skews)

    R = (
        fac1[:, None, None] * skews
        # pyre-fixme[16]: `float` has no attribute `__getitem__`.
        + fac2[:, None, None] * skews_square
        + torch.eye(3, dtype=log_rot.dtype, device=log_rot.device)[None]
    )

    return R, rot_angles, skews, skews_square


def so3_log_map(R: Tensor, eps: float = 0.0001, cos_bound: float = 1e-4) -> Tensor:
    """
    Convert a batch of 3x3 rotation matrices `R`
    to a batch of 3-dimensional matrix logarithms of rotation matrices
    The conversion has a singularity around `(R=I)` which is handled
    by clamping controlled with the `eps` and `cos_bound` arguments.

    Args:
        R: batch of rotation matrices of shape `(minibatch, 3, 3)`.
        eps: A float constant handling the conversion singularity.
        cos_bound: Clamps the cosine of the rotation angle to
            [-1 + cos_bound, 1 - cos_bound] to avoid non-finite outputs/gradients
            of the `acos` call when computing `so3_rotation_angle`.
            Note that the non-finite outputs/gradients are returned when
            the rotation angle is close to 0 or π.

    Returns
    -------
        Batch of logarithms of input rotation matrices
        of shape `(minibatch, 3)`.

    Raises
    ------
        ValueError if `R` is of incorrect shape.
        ValueError if `R` has an unexpected trace.
    """
    _N, dim1, dim2 = R.shape
    if dim1 != 3 or dim2 != 3:
        msg = "Input has to be a batch of 3x3 Tensors."
        raise ValueError(msg)

    phi = so3_rotation_angle(R, cos_bound=cos_bound, eps=eps)

    phi_sin = torch.sin(phi)

    # We want to avoid a tiny denominator of phi_factor = phi / (2.0 * phi_sin).
    # Hence, for phi_sin.abs() <= 0.5 * eps, we approximate phi_factor with
    # 2nd order Taylor expansion: phi_factor = 0.5 + (1.0 / 12) * phi**2
    phi_factor = torch.empty_like(phi)
    ok_denom = phi_sin.abs() > (0.5 * eps)
    # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and `int`.
    phi_factor[~ok_denom] = 0.5 + (phi[~ok_denom] ** 2) * (1.0 / 12)
    phi_factor[ok_denom] = phi[ok_denom] / (2.0 * phi_sin[ok_denom])

    log_rot_hat = phi_factor[:, None, None] * (R - R.permute(0, 2, 1))

    return hat_inv(log_rot_hat)


def hat_inv(h: Tensor) -> Tensor:
    """
    Compute the inverse Hat operator [1] of a batch of 3x3 matrices.

    Args:
        h: Batch of skew-symmetric matrices of shape `(minibatch, 3, 3)`.

    Returns
    -------
        Batch of 3d vectors of shape `(minibatch, 3, 3)`.

    Raises
    ------
        ValueError if `h` is of incorrect shape.
        ValueError if `h` not skew-symmetric.

    [1] https://en.wikipedia.org/wiki/Hat_operator
    """
    _N, dim1, dim2 = h.shape
    if dim1 != 3 or dim2 != 3:
        msg = "Input has to be a batch of 3x3 Tensors."
        raise ValueError(msg)

    ss_diff = torch.abs(h + h.permute(0, 2, 1)).max()

    HAT_INV_SKEW_SYMMETRIC_TOL = 1e-5
    if float(ss_diff) > HAT_INV_SKEW_SYMMETRIC_TOL:
        msg = "One of input matrices is not skew-symmetric."
        raise ValueError(msg)

    x = h[:, 2, 1]
    y = h[:, 0, 2]
    z = h[:, 1, 0]

    return torch.stack((x, y, z), dim=1)


def hat(v: Tensor) -> Tensor:
    """
    Compute the Hat operator [1] of a batch of 3D vectors.

    Args:
        v: Batch of vectors of shape `(minibatch , 3)`.

    Returns
    -------
        Batch of skew-symmetric matrices of shape
        `(minibatch, 3 , 3)` where each matrix is of the form:
            `[    0  -v_z   v_y ]
             [  v_z     0  -v_x ]
             [ -v_y   v_x     0 ]`

    Raises
    ------
        ValueError if `v` is of incorrect shape.

    [1] https://en.wikipedia.org/wiki/Hat_operator
    """
    N, dim = v.shape
    if dim != 3:
        msg = "Input vectors have to be 3-dimensional."
        raise ValueError(msg)

    h = torch.zeros((N, 3, 3), dtype=v.dtype, device=v.device)

    x, y, z = v.unbind(1)

    h[:, 0, 1] = -z
    h[:, 0, 2] = y
    h[:, 1, 0] = z
    h[:, 1, 2] = -x
    h[:, 2, 0] = -y
    h[:, 2, 1] = x

    return h



from __future__ import annotations

import os
import sys
import types
import typing as T


def calling_module_name(
    frames: int = 0, left: int | None = None, strict: bool = True
) -> str:
    """
    Return the name of the module that called the current function from the perspective
    of the function that calls this function.

    Parameters
    ----------
    frames
        The number of frames to go back in the stack offset by 2, by default 0.
    left
        The number nested modules to return, e.g. when the module name is `a.b.c` and `left=1` the return value would
        be `a.b`. By default `None`, which returns the full module name.
    strict
        If the module is not found, raise an error.
        If `False`, return the name of the module that called the current function is returned.

    Returns
    -------
    The name of the module that called the current function.

    """
    import inspect

    frm = inspect.stack()[2 + frames]
    mod = inspect.getmodule(frm[0])
    if mod is None:
        if strict:
            msg = f"Could not find module for {frm.filename}, which was called from {frm.function}"
            raise ModuleNotFoundError(msg)
        return frm.filename.split(".")[0]

    name = mod.__name__
    if left is not None:
        name = ".".join(name.split(".")[: left + 1])

    return name


def caller_identity() -> tuple[str, tuple[str, int, str]]:  # type: ignore
    """
    Identify the caller by a name and hashable key.

    Returns
    -------
    tuple[str,str]
        Module name of the caller and a hashable key
    """
    frame = sys._getframe(2)
    while frame:
        code = frame.f_code
        if os.path.join("utils", __file__) not in code.co_filename:
            mod_name = frame.f_globals["__name__"]
            if mod_name == "__main__":
                mod_name = "<main>"
            return mod_name, (code.co_filename, frame.f_lineno, code.co_name)
        frame = frame.f_back
    return None


def generate_path(obj: T.Any) -> str:
    """
    Inverse of ``locate()``. Generates the fully qualified name of an object.
    Handles cases where the object is not directly importable, e.g. due to
    nested classes or private modules.

    The generated path is simplified by removing redundant module parts, e.g.
    ``module.submodule._impl.class`` may become ``module.submodule.class`` if
    the later also resolves to the same object.

    Bound methods are supported by inspecting the ``__self__`` attribute.

    Parameters
    ----------
    obj
        The object to generate the path for.

    Returns
    -------
    str
        The fully qualified name of the object.
    """

    def __check(path: str, obj: T.Any) -> bool:
        # Check if the path resolves to the same object
        try:
            check_ok = locate_object(path) is obj
        except ImportError:
            check_ok = False
        return check_ok

    try:
        self = obj.__self__
    except AttributeError:
        self = None

    if self is not None:
        self_path = generate_path(self)
        return f"{self_path}.{obj.__name__}"

    module, qualname = obj.__module__, obj.__qualname__

    # Compress the path to this object, e.g. ``module.submodule._impl.class``
    # may become ``module.submodule.class``, if the later also resolves to the same
    # object. This simplifies the string, and also is less affected by moving the
    # class implementation.
    module_parts = module.split(".")
    for k in range(1, len(module_parts)):
        prefix = ".".join(module_parts[:k])
        path = f"{prefix}.{qualname}"
        if __check(path, obj):
            return path

    # Default to the full path plus qualname
    path = f"{module}.{qualname}"
    if not __check(path, obj):
        msg = f"Cannot generate path for object {obj}!"
        raise ImportError(msg)

    return path


def locate_object(path: str) -> T.Any:
    """
    Dynamically locates and returns an object by its fully qualified name.

    Parameters
    ----------
    name (str):
        The fully qualified name of the object to locate.

    Returns
    -------
    Any:
        The located object.

    Raises
    ------
    ImportError
        If the object cannot be located.
    """
    import importlib
    import pydoc

    obj = pydoc.locate(path)

    # Some cases (e.g. torch.optim.sgd.SGD) not handled correctly
    # by pydoc.locate. Try a private function from hydra.
    if obj is None:
        try:
            from hydra.utils import _locate
        except ImportError as e:
            msg = f"Cannot dynamically locate object {path}!"
            raise ImportError(msg) from e
        else:
            obj = _locate(path)  # it raises if fails
    if path == "":
        msg = "Empty path"
        raise ImportError(msg)
    parts = list(path.split("."))
    for part in parts:
        if not len(part):
            raise ValueError(
                f"Error loading '{path}': invalid dotstring."
                + "\nRelative imports are not supported."
            )
    assert len(parts) > 0
    part0 = parts[0]
    try:
        obj = importlib.import_module(part0)
    except Exception as exc_import:
        raise ImportError(
            f"Error loading '{path}':\n{exc_import!r}"
            + f"\nAre you sure that module '{part0}' is installed?"
        ) from exc_import
    for m in range(1, len(parts)):
        part = parts[m]
        try:
            obj = getattr(obj, part)
        except AttributeError as exc_attr:
            parent_dotpath = ".".join(parts[:m])
            if isinstance(obj, types.ModuleType):
                mod = ".".join(parts[: m + 1])
                try:
                    obj = importlib.import_module(mod)
                    continue
                except ModuleNotFoundError as exc_import:
                    raise ImportError(
                        f"Error loading '{path}':\n{exc_import!r}"
                        + f"\nAre you sure that '{part}' is importable from module '{parent_dotpath}'?"
                    ) from exc_import
                except Exception as exc_import:
                    msg = f"Error loading '{path}':\n{exc_import!r}"
                    raise ImportError(msg) from exc_import
            raise ImportError(
                f"Error loading '{path}':\n{exc_attr!r}"
                + f"\nAre you sure that '{part}' is an attribute of '{parent_dotpath}'?"
            ) from exc_attr
    return obj



"""Implements common functional operations for losses."""

from __future__ import annotations

import torch
from torch import Tensor


def split_into_patches(
    x: Tensor,
    sizes: tuple[int, int],
    strides: tuple[int, int] | None = None,
) -> Tensor:
    r"""Splits tensor into N, p_height x p_width blocks."""
    if strides is None:
        strides = (sizes[0], sizes[1])

    # Parse shape as [..., H, W]
    *dots, _, _ = x.shape
    for dim, (size, stride) in enumerate(zip(sizes, strides, strict=False)):
        x = x.unfold(len(dots) + dim, size, stride)

    # ... x P x Hp x Wp
    return x.reshape(*dots, -1, sizes[0], sizes[1])


def reduce_batch(image_loss: Tensor, mask: Tensor) -> Tensor:
    divisor = torch.sum(mask)
    if divisor == 0:
        return 0
    return torch.sum(image_loss) / divisor


def reduce_batch(image_loss: Tensor, mask: Tensor) -> Tensor:
    valid = mask.nonzero()
    image_loss[valid] = image_loss[valid] / mask[valid]

    return torch.mean(image_loss)



"""Operations for embedding data into a (usually higher dimensional) space."""

import enum as E
import math
import typing

import torch


class FourierEmbedMode(E.StrEnum):
    """Various modes of generating Fourier embeddings."""

    SIN = E.auto()
    SIN_COS = E.auto()
    COS = E.auto()
    COS_SIN = E.auto()


type FourierEmbedType = (
    FourierEmbedMode | typing.Literal["sin", "cos", "sin_cos", "cos_sin"]
)


def fourier_embed_2d(
    input: torch.Tensor,
    dim_embed: int,
    *,
    f_max: float | None = None,
    mode: FourierEmbedType = FourierEmbedMode.SIN,
    linear: bool = False,
) -> torch.Tensor:
    """Embed dense 2D data using a Fourier basis."""
    B, C, H, W = input.shape
    factory_kwargs = {"device": input.device, "dtype": input.dtype}
    num_bands = (
        round(dim_embed / (2 * C))
        if mode in (FourierEmbedMode.SIN_COS, FourierEmbedMode.COS_SIN)
        else round(dim_embed / C)
    )
    f_max = max(H, W) / 2.0 if f_max is None else float(f_max)

    if not linear:
        s = 2 ** torch.linspace(
            0.0,
            math.log2(f_max),
            num_bands,
            **factory_kwargs,
        )
    else:
        s = torch.linspace(
            1.0,
            f_max / 2,
            num_bands,
            **factory_kwargs,
        )
    s = s * torch.pi

    input = torch.einsum("bchw,s->bcshw", input, s).reshape(B, -1, H, W)
    match mode:
        case FourierEmbedMode.SIN:
            input = input.sin()
        case FourierEmbedMode.COS:
            input = input.cos()
        case FourierEmbedMode.SIN_COS:
            input = torch.cat([input.sin(), input.cos()], dim=1)
        case FourierEmbedMode.COS_SIN:
            input = torch.cat([input.cos(), input.sin()], dim=1)
        case _:
            msg = f"Unknown Fourier embedding mode: {mode}"
            raise NotImplementedError(msg)

    npad = dim_embed - input.size(1)
    return torch.nn.functional.pad(
        input, (0, 0, 0, 0, 0, npad), mode="constant", value=0
    )



"""Defines commin typings used throughout all submodules."""

import datetime
import os
import pathlib
import typing

import torch
import torch.nn
import torch.types

type Tensor = torch.Tensor
type Device = torch.device | torch.types.Device | typing.Literal["cpu", "cuda"]
type DType = torch.dtype


class TensorFactoryKWArgs(typing.TypedDict, total=False):
    """Keyword arguments for tensor factory functions."""

    device: Device
    dtype: DType
    pin_memory: bool


type StateDict = dict[str, Tensor]
type Size = torch.Size | tuple[int, ...]
type Buffer = bytes | bytearray | memoryview
type Pathable = str | pathlib.Path | os.PathLike
type Primitive = int | float | str | bytes | bytearray | memoryview
type Datetime = datetime.datetime



"""Utilities for contrastive loss functions."""

from __future__ import annotations

import torch
from torch import nn

from unipercept.nn.losses.mixins import ScaledLossMixin


class TripletMarginSimilarityLoss(ScaledLossMixin, nn.TripletMarginWithDistanceLoss):
    """
    Implements a triplet contrastive loss using cosine distance between features, i.e. promoting that the target and
    anchor have a higher cosine similarity than the target and negative.
    """

    def __init__(self, *, margin: float = 0.1, **kwargs) -> None:
        super().__init__(**kwargs, margin=margin, distance_function=cosine_distance)

    def forward(
        self, anchor: torch.Tensor, positive: torch.Tensor, negative: torch.Tensor
    ) -> torch.Tensor:
        loss = nn.functional.triplet_margin_with_distance_loss(
            anchor,
            positive,
            negative,
            distance_function=self.distance_function,
            margin=self.margin,
            swap=self.swap,
            reduction=self.reduction,
        )
        return self.scale * loss


@torch.jit.script
def cosine_distance(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return 1 - nn.functional.cosine_similarity(x, y)



import typing as T

import torch
import torch.compiler
import torch.fx
from torch import nn

from unipercept.types import Tensor
from unipercept.utils.inspect import locate_object

type ActivationFactory = T.Callable[[], nn.Module]
type ActivationSpec = ActivationFactory | type[nn.Module] | nn.Module | None | str


def get_activation(spec: ActivationSpec) -> nn.Module:
    """
    Resolve an activation module from a string, a factory function or a module instance.

    Parameters
    ----------
    activation
        A string, a factory function or an instance of an activation module.
    """
    # Check if no activation is desired
    if spec is None:
        return nn.Identity()

    # Check whether activation is provided as a string or module path
    if isinstance(spec, str):
        spec, *spec_args = spec.split(":", 1) if ":" in spec else (spec, None)
        if len(spec_args) > 0:
            args = map(str.strip, T.cast(list[str], spec_args))
        else:
            args = iter([])
        match spec.lower().strip():
            case "relu":
                return nn.ReLU(
                    inplace=next(args, "").lower() == "inplace",
                )
            case "relu6":
                return nn.ReLU6(
                    inplace=next(args, "").lower() == "inplace",
                )
            case "leaky-relu":
                return nn.LeakyReLU(
                    negative_slope=float(next(args, "0.01")),
                    inplace=next(args, "").lower() == "inplace",
                )
            case "gelu":
                return nn.GELU(
                    approximate=next(args, "none").lower(),
                )
            case "prelu":
                return PReLU(
                    num_parameters=int(next(args, "1")),
                    init=float(next(args, "0.25")),
                )
            case "silu":
                return nn.SiLU()
            case "swish":
                return nn.SiLU()
            case "mish":
                return nn.Mish()
            case "sigmoid":
                return nn.Sigmoid()
            case "tanh":
                return nn.Tanh()
            case "softmax":
                return nn.Softmax(
                    dim=int(next(args, "-1")),
                )
            case "softplus":
                return nn.Softplus()
            case "softsign":
                return nn.Softsign()
            case "threshold":
                if len(spec_args) < 2:
                    msg = (
                        r"Threshold activation requires at least two arguments. "
                        r"Provide them as `theshold:a:b` where `a` is the thehsold and `b` is the value."
                    )
                    raise ValueError(msg)
                return nn.Threshold(
                    threshold=float(next(args)),
                    value=float(next(args)),
                    inplace=next(args, "").lower() == "inplace",
                )
            case "identity":
                return nn.Identity()
            case "none":
                return nn.Identity()
            case _:
                spec = locate_object(spec)

    # If already a module instance, return that instance directly
    if isinstance(spec, nn.Module):
        return spec
    if callable(spec):
        return spec()
    msg = f"Cannot resolve value as an activation module: {spec}"
    raise ValueError(msg)


class ApproxTanhGELU(nn.Module):
    r"""
    An approximate verison of the GELU defined as
    :math:`GELU(x) = \frac{x}{2}(tanh(\sqrt{2/\pi}(x+0.044715x^3)))`.
    """

    @T.override
    def forward(self, x: Tensor) -> Tensor:
        return nn.functional.gelu(x, approximate="tanh")


class ApproxSigmoidGELU(nn.Module):
    r"""
    An approximate verison of the GELU defined as
    :math:`GELU(x) = x\sigma(1.702x)`.
    """

    @T.override
    def forward(self, x: Tensor) -> Tensor:
        return x * (x * 1.702).sigmoid()


class InplaceReLU(nn.Module):
    """A ReLU activation function that performs the operation in-place."""

    @T.override
    def forward(self, x: Tensor) -> Tensor:
        inplace = not (
            torch.jit.is_scripting()
            or torch.compiler.is_compiling()
            or torch.compiler.is_dynamo_compiling()
        )
        return nn.functional.relu(x, inplace=inplace)


class InplaceReLU6(nn.Module):
    """A ReLU6 activation function that performs the operation in-place."""

    @T.override
    def forward(self, x: Tensor) -> Tensor:
        inplace = not (
            torch.jit.is_scripting()
            or torch.compiler.is_compiling()
            or torch.compiler.is_dynamo_compiling()
        )
        return nn.functional.hardtanh(x, 0.0, 6.0, inplace=inplace)


ReLU = nn.ReLU
ReLU6 = nn.ReLU6
LeakyReLU = nn.LeakyReLU
GELU = nn.GELU
SiLU = nn.SiLU
Mish = nn.Mish
Sigmoid = nn.Sigmoid
Tanh = nn.Tanh
Softmax = nn.Softmax
Softsign = nn.Softsign
Threshold = nn.Threshold
Identity = nn.Identity


class PReLU(nn.Module):
    """
    Parametric ReLU activation function. This wraps the :func:`torch.nn.PReLU` module,
    but renames the ``weight`` parameter to ``param`` such that the optimizer can
    differentiate between the network weights that are learned with decay and
    the second-class parameters that are learned without decay.
    """

    __constants__ = ["num_parameters"]
    num_parameters: int

    def __init__(
        self, num_parameters: int = 1, init: float = 0.25, device=None, dtype=None
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        self.num_parameters = num_parameters
        super().__init__()
        self.init = init
        self.param = nn.Parameter(torch.empty(num_parameters, **factory_kwargs))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        torch.nn.init.constant_(self.param, self.init)

    def forward(self, input: Tensor) -> Tensor:
        return nn.functional.prelu(input, self.param)

    def extra_repr(self) -> str:
        return f"num_parameters={self.num_parameters}"


############
# Softplus #
############

_SOFTPLUS_DEFAULT_BETA: T.Final[float] = 1.0
_SOFTPLUS_DEFAULT_THRESHOLD: T.Final[float] = 20.0
_SOFTPLUS_DEFAULT_EPS: T.Final[float] = torch.finfo(torch.float32).resolution


def softplus(
    x: Tensor,
    beta: float = _SOFTPLUS_DEFAULT_BETA,
    threshold: float = _SOFTPLUS_DEFAULT_THRESHOLD,
) -> Tensor:
    """
    Softplus function with defaults that match the inverse defined in this module.

    Parameters
    ----------
    x
        Input tensor.
    beta
        Coefficient for the softplus function.
    threshold
        Threshold for the softplus function. The result is linear for (x * beta) > threshold.

    Returns
    -------
    Tensor
        Output tensor.
    """
    return nn.functional.softplus(x, beta=beta, threshold=threshold)


class Softplus(nn.Module):
    """Numerically stable softplus function. See :func:`softplus`."""

    def __init__(
        self,
        beta: float = _SOFTPLUS_DEFAULT_BETA,
        threshold: float = _SOFTPLUS_DEFAULT_THRESHOLD,
    ) -> None:
        super().__init__()
        self.beta = beta
        self.threshold = threshold

    @T.override
    def forward(self, x: Tensor) -> Tensor:
        return softplus(x, beta=self.beta, threshold=self.threshold)


def inverse_softplus(
    y: Tensor,
    beta: float = _SOFTPLUS_DEFAULT_BETA,
    threshold: float = _SOFTPLUS_DEFAULT_THRESHOLD,
    eps=1e-6,
) -> Tensor:
    """
    Numerically stable inverse of the softplus function.

    Note that this can only approximate the inverse of values that
    are greater than zero, as the softplus function is not invertible
    for negative values.

    Parameters
    ----------
    y
        Input tensor (output of the softplus function).
    beta
        Coefficient for the softplus function.
    threshold
        Numerical-stability threshold applied to ``y * beta`` in the inverse
        computation.
    eps
        Small constant for numerical stability in edge cases.

    Returns
    -------
    Tensor
        Output tensor.
    """
    return y.where(y * beta >= threshold, y + (-(-y * beta).expm1()).log() / beta)


class InverseSoftplus(nn.Module):
    """Numerically stable inverse of the softplus function. See :func:`inverse_softplus`."""

    def __init__(
        self,
        beta: float = _SOFTPLUS_DEFAULT_BETA,
        threshold: float = _SOFTPLUS_DEFAULT_THRESHOLD,
        eps: float = _SOFTPLUS_DEFAULT_EPS,
    ) -> None:
        super().__init__()
        self.beta = beta
        self.threshold = threshold
        self.eps = eps

    @T.override
    def forward(self, x: Tensor) -> Tensor:
        return inverse_softplus(
            x, beta=self.beta, threshold=self.threshold, eps=self.eps
        )



r"""
Implementation of a context manager to measure flops of a module.

Inspired by the discussion on [the ideal flop counter](https://dev-discuss.pytorch.org/t/the-ideal-pytorch-flop-counter-with-torch-dispatch/505)
on the PyTorch forum.
"""

import logging
import operator
import typing as T
from collections import defaultdict
from functools import reduce

import torch
from torch.utils._python_dispatch import TorchDispatchMode
from torch.utils._pytree import PyTree, tree_map

aten: torch._ops._OpNamespace = torch.ops.aten
type InputType = torch.Tensor | bool | tuple[bool, ...]


def _matmul_flop_count(
    inputs: tuple[torch.Tensor, ...], _outputs: tuple[T.Any, ...]
) -> int:
    """Count flops for matmul."""
    # Inputs should be a list of length 2.
    # Inputs contains the shapes of two matrices.
    input_shapes = [v.shape for v in inputs]
    assert len(input_shapes) == 2, input_shapes
    assert input_shapes[0][-1] == input_shapes[1][-2], input_shapes
    return inputs[0].numel() * input_shapes[-1][-1]


def _addmm_flop_count(inputs: tuple[torch.Tensor], _outputs: tuple[T.Any, ...]) -> int:
    """Count flops for fully connected layers."""
    # Count flop for nn.Linear
    # inputs is a list of length 3.
    input_shapes = [v.shape for v in inputs[1:3]]
    # input_shapes[0]: [batch size, input feature dimension]
    # input_shapes[1]: [batch size, output feature dimension]
    assert len(input_shapes[0]) == 2, input_shapes[0]
    assert len(input_shapes[1]) == 2, input_shapes[1]
    batch_size, input_dim = input_shapes[0]
    output_dim = input_shapes[1][1]
    return batch_size * input_dim * output_dim


def _bmm_flop_count(inputs: tuple[torch.Tensor], _outputs: tuple[T.Any, ...]) -> int:
    """Count flops for the bmm operation."""
    # Inputs should be a list of length 2.
    # Inputs contains the shapes of two tensor.
    assert len(inputs) == 2, len(inputs)
    input_shapes = [v.shape for v in inputs]
    n, c, t = input_shapes[0]
    d = input_shapes[-1][-1]
    return n * c * t * d


def _conv_flop_count(
    x_shape: torch.Size | list[int],
    w_shape: torch.Size | list[int],
    out_shape: torch.Size | list[int],
    transposed: bool = False,
) -> int:
    """
    Count flops for convolution. Note only multiplication is
    counted. Computation for addition and bias is ignored.
    Flops for a transposed convolution are calculated as.

    ::math
        flops = (x_shape[2:] * prod(w_shape) * batch_size).

    Parameters
    ----------
        x_shape (list(int)): The input shape before convolution.
        w_shape (list(int)): The filter shape.
        out_shape (list(int)): The output shape after convolution.
        transposed (bool): is the convolution transposed
    Returns:
        int: the number of flops
    """
    batch_size = x_shape[0]
    conv_shape = (x_shape if transposed else out_shape)[2:]
    return (
        batch_size
        * reduce(operator.mul, w_shape, 1)
        * reduce(operator.mul, conv_shape, 1)
    )


def _conv_flop_count(
    inputs: list[torch.Tensor | bool | tuple[bool]],
    outputs: tuple[torch.Tensor],
) -> int:
    """Count flops for convolution."""
    x: torch.Tensor = T.cast(torch.Tensor, inputs[0])
    w: torch.Tensor = T.cast(torch.Tensor, inputs[1])
    x_shape, w_shape, out_shape = (x.shape, w.shape, outputs[0].shape)
    transposed: bool = T.cast(bool, inputs[6])

    return _conv_flop_count(
        list(x_shape), list(w_shape), list(out_shape), transposed=transposed
    )


def _transpose_shape(shape: torch.Size) -> list[int]:
    return [shape[1], shape[0], *list(shape[2:])]


def _conv_backward_flop_count(
    inputs: tuple[torch.Tensor | bool | tuple[bool]], outputs: tuple[torch.Tensor]
) -> int:
    grad_out_shape, x_shape, w_shape = (
        T.cast(torch.Tensor, i).shape for i in inputs[:3]
    )
    output_mask = inputs[-1]
    fwd_transposed = inputs[7]
    flop_count: int = 0

    if T.cast(tuple[bool], output_mask)[0]:
        grad_input_shape = outputs[0].shape

        flop_count = flop_count + _conv_flop_count(
            grad_out_shape, w_shape, grad_input_shape, not fwd_transposed
        )
    if T.cast(tuple[bool], output_mask)[1]:
        grad_weight_shape = outputs[1].shape

        flop_count += _conv_flop_count(
            list(_transpose_shape(x_shape)),
            list(grad_out_shape),
            list(grad_weight_shape),
            T.cast(bool, fwd_transposed),
        )

    return flop_count


flop_mapping: T.Mapping[
    T.Callable[..., T.Any], T.Callable[[tuple[T.Any, ...], tuple[T.Any, ...]], int]
] = {
    aten.mm: _matmul_flop_count,
    aten.matmul: _matmul_flop_count,
    aten.addmm: _addmm_flop_count,
    aten.bmm: _bmm_flop_count,
    aten.convolution: _conv_flop_count,
    aten._convolution: _conv_flop_count,
    aten.convolution_backward: _conv_backward_flop_count,
    # Add their default to make sure they can be mapped.
    aten.mm.default: _matmul_flop_count,
    aten.matmul.default: _matmul_flop_count,
    aten.addmm.default: _addmm_flop_count,
    aten.bmm.default: _bmm_flop_count,
    aten.convolution.default: _conv_flop_count,
    aten._convolution.default: _conv_flop_count,
    aten.convolution_backward.default: _conv_backward_flop_count,
}


def _normalize_tuple[_T](x: _T) -> tuple[_T]:
    if not isinstance(x, tuple):
        return (x,)
    return x


class FlopTensorDispatchMode(TorchDispatchMode):
    """
    A context manager to measure flops of a module. Requires PyTorch 1.13+.

    Examples::

        >>> import copy
        >>> import torch
        >>> import torchvision.models as models
        >>> from torchtnt.utils.flops import FlopTensorDispatchMode

        >>> module = models.resnet18()
        >>> module_input = torch.randn(1, 3, 224, 224)
        >>> with FlopTensorDispatchMode(module) as ftdm:
        >>>     # count forward flops
        >>>     res = module(module_input).mean()
        >>>     flops_forward = copy.deepcopy(ftdm.flop_counts)

        >>>     # reset count before counting backward flops
        >>>     ftdm.reset()
        >>>     res.backward()
        >>>     flops_backward = copy.deepcopy(ftdm.flop_counts)

    """

    def __init__(self, module: torch.nn.Module) -> None:
        """
        Initializes a FlopTensorDispatchMode context manager object.

        Parameters
        ----------
        module
            The module to count flops on.
        """
        self._all_hooks: list[torch.utils.hooks.RemovableHandle] = []
        self._instrument_module(module, "")

        self.flop_counts: defaultdict[str, defaultdict[str, int]] = defaultdict(
            lambda: defaultdict(int)
        )
        self._parents: list[str] = [""]

    def __exit__(self, exc_type, exc_val, exc_tb):
        for hook_handle in self._all_hooks:
            hook_handle.remove()
        super().__exit__(exc_type, exc_val, exc_tb)

    def __torch_dispatch__(
        self,
    ) -> PyTree:
        rs = func(*args, **kwargs)
        outs = _normalize_tuple(rs)

        if func in flop_mapping:
            flop_count = flop_mapping[func](args, outs)
            for par in self._parents:
                self.flop_counts[par][func.__name__] += flop_count
        else:
            logging.debug(f"{func} is not yet supported in FLOPs calculation.")

        return rs

    def _create_backwards_push(self, name: str) -> T.Callable[..., T.Any]:
        class PushState(torch.autograd.Function):
            @staticmethod
            def forward(ctx, *args):
                args = tree_map(
                    lambda x: x.clone() if isinstance(x, torch.Tensor) else x, args
                )
                if len(args) == 1:
                    return args[0]
                return args

            @staticmethod
            def backward(ctx, *grad_outs):
                parents = self._parents
                parents.append(name)
                return grad_outs

        # But this class can't be lifted out of the function as it is a static class
        # using a function parameter.
        return PushState.apply

    def _create_backwards_pop(self, name: str) -> T.Callable[..., T.Any]:
        class PopState(torch.autograd.Function):
            @staticmethod
            def forward(ctx, *args):
                args = tree_map(
                    lambda x: x.clone() if isinstance(x, torch.Tensor) else x, args
                )
                if len(args) == 1:
                    return args[0]
                return args

            @staticmethod
            def backward(ctx, *grad_outs):
                parents = self._parents
                assert parents[-1] == name
                parents.pop()
                return grad_outs

        # But this class can't be lifted out of the function as it is a static class
        # using a function parameter.
        return PopState.apply

    def _enter_module(self, name: str) -> T.Callable[..., T.Any]:
        def f(module: torch.nn.Module, inputs: tuple[T.Any, ...]):
            parents = self._parents
            parents.append(name)
            inputs = _normalize_tuple(inputs)
            return self._create_backwards_pop(name)(*inputs)

        return f

    def _exit_module(self, name: str) -> T.Callable[..., T.Any]:
        def f(
            module: torch.nn.Module,
            inputs: tuple[T.Any, ...],
            outputs: tuple[T.Any, ...],
        ):
            parents = self._parents
            assert parents[-1] == name
            parents.pop()
            outputs = _normalize_tuple(outputs)
            return self._create_backwards_push(name)(*outputs)

        return f

    def _instrument_module(
        self,
        mod: torch.nn.Module,
        par_name: str,
    ) -> None:
        for name, module in dict(mod.named_children()).items():
            formatted_name = name
            if par_name != "":
                formatted_name = f"{par_name}.{name}"
            self._all_hooks.append(
                module.register_forward_pre_hook(self._enter_module(formatted_name))
            )
            self._all_hooks.append(
                module.register_forward_hook(self._exit_module(formatted_name))
            )
            self._instrument_module(module, formatted_name)

    def reset(self) -> None:
        """Resets current flop count."""
        self._parents = [""]
        self.flop_counts.clear()



import functools
import inspect
import pickle
import typing as T
import warnings
from typing import override

import cloudpickle

__all__ = ["as_picklable", "pickles"]


_LAMBDA_NAME: T.Final = (lambda: None).__name__
_LOCALS_NAME: T.Final = "<locals>"
_ALWAYS_PICKLES: T.Final[tuple[type, ...]] = (functools.partial,)


def _has_name(obj: T.Callable, name: str) -> bool:
    """Check if a function has a name."""
    try:  # Check for __name__ attribute, if it exists
        privnaem = getattr(obj, "__name__", None)
        if privnaem is not None and name in privnaem:
            return True
    except AttributeError:
        pass  # FIXME! wrong attribute?

    try:  # Check for __qualname__ attribute, if it exists
        qualname = getattr(obj, "__qualname__", None)
        if qualname is not None and name in qualname:
            return True
    except AttributeError:
        pass  # FIXME!

    return False


def _is_safe_class(cls: type) -> bool:
    """Check if a class is a pickling class."""
    if not callable(cls):
        return False
    if not inspect.isclass(cls):
        return False
    return isinstance(cls, _ALWAYS_PICKLES)


def _is_lambda(fn: T.Callable) -> bool:
    """Check if a function is a lambda."""
    return _has_name(fn, _LAMBDA_NAME)


def _is_closure(fn: T.Callable) -> bool:
    """Applies some heuristics on the function/class name to check if it is a local function, and thus a closure."""
    return _has_name(fn, _LOCALS_NAME)


def pickles(fn: T.Callable, use_experiment=True) -> bool:
    """Check if a function can be pickled, uses various heuristics."""
    if not callable(fn):
        return False

    if _is_lambda(fn) or _is_closure(fn):
        return False

    if _is_safe_class(fn):
        return True

    name = getattr(fn, "__name__", None)
    if use_experiment:
        try:
            pickle.dumps(fn)
            return True
        except (Exception, pickle.PicklingError):
            warnings.warn(
                f"Cannot pickle {name} using direct pickle experiment.",
                stacklevel=2,
            )
            return False
    else:
        warnings.warn(
            "Cannot determine whether {name} is picklable using heuristics, assuming it is not.",
            stacklevel=2,
        )

    return False


class as_picklable[_T: T.Callable]:
    """A decorator that makes a function picklable, based on Detectron2's implementation ``PicklableWrapper``."""

    __slots__ = ("_fn",)

    def __new__(cls, fn: _T) -> _T:
        if isinstance(fn, cls):
            return T.cast(_T, fn)
        if pickles(fn):
            return fn

        fn_wrap = super().__new__(cls)
        fn_wrap._fn = fn

        return T.cast(_T, fn_wrap)

    @override
    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self._fn))

    @override
    def __eq__(self, other: object) -> bool:
        is_our_class = isinstance(other, self.__class__)

        if is_our_class:
            return self._fn == other._fn
        assert callable(other), f"Expected callable, got {type(other)}"
        return self._fn == other

    @override
    def __str__(self) -> str:
        return str(self._fn)

    @override
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._fn}>"

    @override
    def __reduce__(self):
        s = cloudpickle.dumps(self._fn)
        return cloudpickle.loads, (s,)

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)

    def __getattr__(self, attr):
        if attr == "__slots__" or attr in self.__slots__:
            return super().__getattribute__(attr)
        return self._fn.__getattribute__(attr)

    @override
    def __setattr__(self, attr: T.Never, val: T.Never) -> T.Any:
        if attr == "__slots__" or attr in self.__slots__:
            return super().__setattr__(attr, val)
        return self._fn.__setattr__(attr, val)



"""Implements point sampling and selection methods proposed in PointRend [1]_.

See Also
--------
- `Reference implementation <https://github.com/facebookresearch/detectron2/tree/main/projects/PointRend>`_
    provided in [1]_.

References
----------
[1] `PointRend: Image Segmentation as Rendering <https://arxiv.org/abs/1912.08193>`_
"""

import enum
import math
import typing
from collections.abc import Callable

import torch
from torch.nn.functional import grid_sample

from unipercept.types import Tensor
from unipercept.utils.check import assert_shape


def sample(
    input: Tensor,
    point_coords: Tensor,
    align_corners: bool = False,
    **kwargs: typing.Any,
) -> Tensor:
    """Sample points from a grid/image.

    A wrapper around :function:`torch.nn.functional.grid_sample` to support 3D
    point coordinate tensors.
    Unlike :function:`torch.nn.functional.grid_sample` it assumes `point_coords` to
    lie inside [0, 1] x [0, 1] square.

    Parameters
    ----------
    input: Tensor
        A tensor of shape (N, C, H, W) that contains features map on a H x W grid.
    point_coords: Tensor
        A tensor of shape (N, P, 2) or (N, Hgrid, Wgrid, 2) that contains
        [0, 1] x [0, 1] normalized point coordinates.
    align_corners: bool
        Whether to align the corners of the input and output tensors.

    Returns
    -------
    Tensor
        A tensor of shape (N, C, P) or (N, C, Hgrid, Wgrid) that contains
        features for points in `point_coords`.
        The features are obtained via interpolation from `input` the same way as
        :function:`torch.nn.functional.grid_sample`.

    Examples
    --------
    >>> input = torch.rand(1, 3, 5, 5)
    >>> point_coords = torch.rand(1, 10, 2)
    >>> output = sample(input, point_coords)
    >>> print(output.shape)
    torch.Size([1, 3, 10])
    """
    assert_shape(input, ("B", "C", "H", "W"))
    assert_shape(point_coords, (..., 2))

    add_dim = False
    if point_coords.dim() == 3:
        add_dim = True
        point_coords = point_coords.unsqueeze(2)
    output = grid_sample(
        input,
        2.0 * point_coords - 1.0,
        align_corners=align_corners,
        padding_mode="border",
        **kwargs,
    )
    if add_dim:
        output = output.squeeze(3)
    return output


@torch.no_grad()
def logit_uncertainty(logits: Tensor) -> Tensor:
    """
    Method proposed in PointRend [1]_ to calculates the uncertainty of logits.

    See :function:`random_points_with_importance`, where this is the default importance
    function.

    Parameters
    ----------
    logits: Tensor
        Input tensor of shape (N, C, H, W) or (N, H, W).

    Returns
    -------
    Tensor
        Uncertainty tensor of the same shape as the input.

    Examples
    --------
    >>> logits = torch.tensor([[-1.0, 0.5], [0.2, -0.3]])
    >>> uncertainty = logit_uncertainty(logits)
    >>> print(uncertainty)
    tensor([[-1.0000, -0.5000],
            [-0.2000, -0.3000]])
    """
    return -logits.detach().abs()


def random_points(
    source: Tensor,
    n_points: int,
    d_coord: int = 2,
    mask: Tensor | None = None,
) -> Tensor:
    """
    Sample points in [0, 1] x [0, 1] coordinate space.

    The points are sampled uniformly can can optionally be conditioned on a mask that
    defines the region of interest.

    Parameters
    ----------
    source: Tensor
        A tensor of shape $(N, C, ...)$
    n_points: int
        The number of points $P$ to sample.
    d_coord: int
        The number of dimensions $S$ of the coordinate space, by default 2d.
    mask: Tensor, optional
        A boolean tensor of shape $(N, 1, ...)$ that contains a mask of the region of
        interest. Random points will only be selected such that they lie inside the
        mask.

    Returns
    -------
    Tensor
        A tensor of shape $(N, P, S)$ that contains $S$d coordinates of $P$ points.

    Examples
    --------
    >>> source = torch.rand(1, 1, 5, 5)
    >>> points = random_points(source, n_points=10)
    >>> print(points.shape)
    torch.Size([1, 10, 2])
    """
    assert_shape(source, ("B", "C", *(f"C_{i}" for i in range(d_coord))))

    n_batch = source.size(0)

    # Sample random coordinates in [0, 1] for `d_coord` spatial dimensions.
    point_coords = torch.rand(
        n_batch, n_points, d_coord, dtype=torch.float32, device=source.device
    )

    # No mask constaints, return the coordinates directly.
    if mask is None:
        return point_coords

    # Currently only 2D masks are supported
    if d_coord != 2:
        msg = "Mask constraints are only supported for 2D coordinates."
        raise NotImplementedError(msg)

    # Ensure the sampled points lie inside the mask by sampling random pixel coordinates
    # from the mask and adding the point coordinates normalized by the spatial size
    # as additive noise
    mask_flat = mask.view(n_batch, -1)
    mask_cumsum = mask_flat.cumsum(dim=1)

    # Generate uniform samples from 0 to the max of the cumsum
    max_cumsum = mask_cumsum[:, -1].unsqueeze(1)
    random_samples = torch.rand(n_batch, n_points, device=source.device) * max_cumsum

    # Find the indices in the cumulative sum that correspond to each random sample
    idx = torch.searchsorted(mask_cumsum, random_samples)

    # Convert linear indices to subscripts
    idx_subscripts = torch.stack(
        (
            torch.div(idx, mask.shape[-1], rounding_mode="trunc"),
            idx % mask.shape[-1],
        ),
        dim=-1,
    )

    # Convert indices to pixel (center) coordinates
    idx_subscripts = idx_subscripts.float() + 0.5
    point_coords = point_coords - 0.5

    # Convert the pixel indices to normalized coordinates
    spatial_size = torch.tensor(
        mask.shape[-d_coord:], dtype=torch.float32, device=source.device
    )

    # Flip X and Y coordinates
    idx_subscripts = idx_subscripts.flip(-1)
    spatial_size = spatial_size.flip(-1)

    # Normalize and add uniform noise within each pixel
    pixel_coords = idx_subscripts.float() / spatial_size
    point_noise = point_coords / spatial_size

    return pixel_coords + point_noise


def random_points_with_importance(
    source: Tensor,
    n_points: int,
    d_coord: int = 2,
    mask: Tensor | None = None,
    oversample_ratio: float = 3,
    importance_fn: Callable[[Tensor], Tensor] = logit_uncertainty,
    importance_sample_ratio: float = 0.75,
    **kwargs,
) -> Tensor:
    """
    Sample points in $[0, 1]$ x [0, 1]$ coordinate space based on their importance,
    which PointRend [1]_ uses to supervise imports based on their uncertainty.

    Notes
    -----
    It is crucial to calculate uncertainty based on the sampled prediction value for
    the points.
    Calculating uncertainties of the coarse predictions first and sampling them for
    points leads
    to incorrect results.

    To illustrate this: assume importance_fn(src)=-abs(src), a sampled point between
    two coarse predictions with -1 and 1 logits has 0 logits, and therefore 0
    uncertainty value.
    However, if we calculate uncertainties for the coarse predictions first,
    both will have -1 uncertainty, and the sampled point will get -1 uncertainty.

    Parameters
    ----------
    source: Tensor
        A tensor of shape $(N, C, ...)$
    n_points: int
        The number of points $P$ to sample.
    d_coord: int
        The number of dimensions $S$ of the coordinate space, by default 2d.
    mask: Tensor, optional
        A boolean tensor of shape $(N, 1, ...)$ that contains a mask of the region of
        interest.
        Random points will only be selected such that they lie inside the mask.
    oversample_ratio: int
        Oversampling parameter, controls the amount of points $P_s$ over which
        the importance is computed.
    importance_sample_ratio: float
        Ratio of points that are sampled via importance sampling.
    importance_fn: Callable[[Tensor], Tensor]
        A function that takes samples points of shape $(N, C, P_s)$ that returns their
        importance as $(N, 1, P_s)$.
    **kwargs:
        Additional arguments for :function:`torch.nn.functional.grid_sample`.

    Returns
    -------
    Tensor
        A tensor of shape $(N, P, S)$ that contains $S$d coordinates of $P$ points.

    Examples
    --------
    >>> source = torch.rand(1, 1, 5, 5)
    >>> points = random_points_with_importance(source, n_points=10)
    >>> print(points.shape)
    torch.Size([1, 10, 2])
    """
    assert oversample_ratio >= 1
    assert importance_sample_ratio <= 1
    assert importance_sample_ratio >= 0

    n_batch = source.shape[0]
    n_sample = int(n_points * oversample_ratio)

    point_coords = random_points(source, n_sample, d_coord, mask)
    samples = sample(source, point_coords, **kwargs)
    point_importance = importance_fn(samples)

    # Select the most important points
    n_important = int(importance_sample_ratio * n_points)
    i_important = torch.topk(point_importance[:, 0, :], k=n_important, dim=1)[1]

    shift = n_sample * torch.arange(n_batch, dtype=torch.long, device=source.device)
    i_important += shift[:, None]

    point_coords = point_coords.view(-1, 2)[i_important.view(-1), :].view(
        n_batch, n_important, d_coord
    )

    # Fill the rest with random points to ensure a total of `n_points`
    n_random = n_points - n_important
    if n_random > 0:
        point_coords = torch.cat(
            (point_coords, random_points(source, n_random, d_coord, mask)),
            dim=1,
        )

    # The result is of shape (n_batch, n_points, 2)
    return point_coords


class BinsMode(enum.StrEnum):
    """An enumeration that defines the mode of binning."""

    QUANTILE = enum.auto()
    LINEAR = enum.auto()


@torch.no_grad()
def bins_by_values(
    source: Tensor,
    n_bins: int,
    mask: Tensor | None = None,
    use_batch: bool = False,
    mode: BinsMode | typing.Literal["linear", "quantile"] = BinsMode.QUANTILE,
    eps: float = 1e-6,
) -> Tensor:
    r"""
    Create linear bins based on quantiles of the source tensor.

    Parameters
    ----------
    source: Tensor[B, C, *]
        A tensor of shape $(B, ...)$
    n_bins: int
        The number of bins $N$ to create.
    mask: Tensor[B, *], optional
        A mask of shape $(B, ...)$ that defines the valid region of the source tensor.
    use_batch: bool
        If set to `True`, the bins are computed based on the per-batch values. Otherwise
        the bins are computed based on the global values.
    mode: BinsMode or str
        Mode of binning, either "quantile" or "linear".
    eps: float
        Small value to avoid division by zero.

    Returns
    -------
    Tensor[B, N]
        Tensor containing the bin boundaries.

    Examples
    --------
    >>> source = torch.rand(1, 1, 5, 5)
    >>> bins = bins_by_values(source, n_bins=3)
    >>> print(bins)
    tensor([[0.3333, 0.6667, 1.0000]])
    """
    assert_shape(source, ("B", "C", ...))

    # Batch dims
    n_batch = source.shape[0]

    # Compute global or per-batch bins?
    if use_batch:
        source_flat = source.flatten(1)
        if mask is not None:
            mask_flat = mask.view_as(source_flat)
            # Use a large value for masked out elements to push them to the end
            # This is a common trick for quantile on masked tensors in a batch
            # because torch.quantile doesn't support masking per-batch easily.
            source_masked = torch.where(
                mask_flat, source_flat, torch.tensor(float("nan"), device=source.device)
            )
        else:
            source_masked = source_flat

        quantiles = torch.linspace(0, 1, n_bins + 1, device=source.device)[1:]
        # vmap-like behavior for quantile is not yet standard, use a loop but make it tighter
        # or use quantile if all rows have same number of valid elements.
        # For simplicity and graph-compatibility, if we can't vectorize, we keep it as is
        # but try to avoid recursion.

        results = []
        for i in range(n_batch):
            row = source_masked[i]
            row_valid = row[~row.isnan()]
            match mode:
                case BinsMode.QUANTILE:
                    results.append(torch.quantile(row_valid, quantiles) + eps)
                case BinsMode.LINEAR:
                    results.append(
                        torch.linspace(
                            row_valid.min(), row_valid.max() + eps, n_bins + 1
                        )[1:]
                    )
        return torch.stack(results, dim=0)

    # Filter out the masked values
    if mask is not None:
        source_valid = source[mask.view_as(source)]
    else:
        source_valid = source[~source.isnan()]

    if source_valid.numel() == 0:
        return torch.zeros((n_batch, n_bins), device=source.device)

    # Compute the bins based on the selected mode
    match mode:
        case BinsMode.QUANTILE:
            quantiles = torch.linspace(0, 1, n_bins + 1, device=source.device)[1:]
            bins = torch.quantile(source_valid, quantiles) + eps
        case BinsMode.LINEAR:
            bins = torch.linspace(
                source_valid.min(), source_valid.max() + eps, n_bins + 1
            )[1:]
        case _:
            msg = f"Invalid mode: {mode}"
            raise ValueError(msg)

    return bins.unsqueeze(0).repeat(n_batch, 1)


def map_in_bins(
    source: Tensor,
    bins: Tensor,
    fn: Callable[[Tensor, Tensor], Tensor],
    mask: Tensor | None = None,
    values: Tensor | None = None,
    **kwargs,
) -> Tensor:
    r"""
    Apply a function to a source tensor by dividing it into ``n_bins`` bins and applying
    the function ``fn`` to each bin.

    The resulting tensor is concatenated along the channel (C) dimension.

    Parameters
    ----------
    source: Tensor[B, C, *]
        A tensor whose values are divided into bins.
    bins: Tensor[B, N]
        A tensor that contains the bin boundaries.
    fn: Callable[[Tensor[B, C, *], Tensor[B, C, *]], Tensor[B, R, *]]
        A function that is applied to each bin of the source tensor. Returns :math:`R`
        channels, which are concatenated into :math:`C \times R` output channels.
        The first argument is the source tensor, and the second argument is a boolean
        mask of the same shape that indicates which values belong to the current bin.
    mask: Tensor[B, *], optional
        Mask for invalid values, which are combined with the binned mask usign a logcal AND.
    values: Tensor[B, *], optional
        A tensor that contains the values to be binned. If not provided, the source tensor
        is used.

    Returns
    -------
    Tensor[B, N * R, *]
        Resulting tensor after applying the function to each bin.

    Examples
    --------
    >>> source = torch.rand(1, 1, 5, 5)
    >>> bins = torch.tensor([[0.2, 0.4, 0.6, 0.8]])
    >>> fn = lambda s, m: s * m
    >>> result = map_in_bins(source, bins, fn)
    >>> print(result.shape)
    torch.Size([1, 4, 5, 5])
    """
    assert_shape(source, ("B", "C", ...))
    assert_shape(bins, ("B", "N"))

    n_bins = bins.shape[1]

    # Create a mask that indicates which bin a value belongs to
    if values is None:
        values = source

    # Flatten values to match bins dimension for searchsorted
    orig_shape = values.shape
    values_flat = values.flatten(1)
    mask_bin = torch.searchsorted(bins, values_flat, right=True)
    mask_bin = mask_bin.view(orig_shape)

    # Apply the function to each bin
    outputs = []
    for i in range(n_bins):
        mask_i = mask_bin == i
        if mask is not None:
            mask_i = mask_i & mask
        if not mask_i.any():
            continue
        output_i = fn(source, mask_i)
        if output_i.ndim == 0:
            output_i = output_i.reshape(1)
        outputs.append(output_i)

    # Concatenate the results along the channel dimension
    if not outputs:
        return torch.empty(0, device=source.device)

    # Use dim=0 if outputs are 1D, otherwise use dim=1
    concat_dim = 1 if outputs[0].ndim > 1 else 0
    return torch.cat(outputs, dim=concat_dim)


def random_points_with_bins(
    source: Tensor,
    n_points: int,
    d_coord: int = 2,
    mask: Tensor | None = None,
    n_bins: int = 10,
    use_batch: bool = False,
    mode: BinsMode | typing.Literal["linear", "quantile"] = BinsMode.QUANTILE,
    values: Tensor | None = None,
) -> Tensor:
    """
    Samples points uniformly from the source tensor, where the points are first divided
    into bins based on their values.

    See :function:`random_points` for more details.

    Parameters
    ----------
    source: Tensor
        A tensor of shape $(N, C, ...)$
    n_points: int
        The number of points $P$ to sample.
    d_coord: int
        The number of dimensions $S$ of the coordinate space, by default 2d.
    mask: Tensor, optional
        A boolean tensor of shape $(N, 1, ...)$ that contains a mask of the region of
        interest. Random points will only be selected such that they lie inside the
        mask.
    n_bins: int
        The number of bins $N$ to create.
    use_batch: bool
        If set to `True`, the bins are computed based on the per-batch values. Otherwise
        the bins are computed based on the global values.
    mode: BinsMode or str
        Mode of binning, either "quantile" or "linear".
    values: Tensor, optional
        A tensor that contains the values to be binned. If not provided, the source tensor
        is used.

    Returns
    -------
    Tensor
        A tensor of shape $(N, P, S)$ that contains $S$d coordinates of $P$ points.

    Examples
    --------
    >>> source = torch.rand(1, 1, 5, 5)
    >>> points = random_points_with_bins(source, n_points=10)
    >>> print(points.shape)
    torch.Size([1, 10, 2])
    """
    if values is None:
        values = source
    if mask is None:
        mask = torch.ones_like(values, dtype=torch.bool)

    bins = bins_by_values(source, n_bins, mask=mask, mode=mode, use_batch=use_batch)

    points = []
    for s, m, b, v in zip(source, mask, bins, values, strict=True):
        s = s.unsqueeze(0)
        m = m.unsqueeze(0)
        b = b.unsqueeze(0)
        v = v.unsqueeze(0)

        points_per_bin = math.ceil(
            n_points
            / map_in_bins(s, b, lambda _, v: v.any(), mask=m, values=v)
            .int()
            .sum()
            .item()
        )

        res = map_in_bins(
            s,
            b,
            lambda s, m: random_points(s, points_per_bin, d_coord, mask=m),
            mask=m,
        ).squeeze(0)
        if res.shape[0] < n_points:
            padding = torch.zeros(
                (n_points - res.shape[0], d_coord), device=res.device, dtype=res.dtype
            )
            res = torch.cat([res, padding], dim=0)
        points.append(res[:n_points, ...])
    return torch.stack(points, dim=0)



"""Basic MLP and variants."""

import contextlib
import typing as T

import torch
from torch import nn

from unipercept.nn.activations import ActivationSpec, get_activation
from unipercept.nn.norms import NormSpec, get_default_bias, get_norm

__all__ = ["MLP"]

EPS = torch.finfo(torch.float32).eps


class MLP(nn.Module):
    """Multi-layer Perception layer."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        hidden_features: float = 1.0,
        *,
        dropout: float | T.Sequence[float] | tuple[float, float] = 0.0,
        layers: int = 2,
        norm: (
            NormSpec
            | T.Sequence[NormSpec]
            | tuple[NormSpec, NormSpec]
            | tuple[NormSpec, NormSpec, NormSpec]
        ) = None,
        bias: tuple[bool, bool] | bool | None = None,
        activation: (
            ActivationSpec
            | T.Sequence[ActivationSpec]
            | tuple[ActivationSpec, ActivationSpec]
            | tuple[ActivationSpec, ActivationSpec, ActivationSpec]
        ) = nn.GELU,
        init_gain: float | None = None,
    ) -> None:
        """
        Parameters
        ----------
        in_features
            Number of input channels.
        out_features
            Number of output channels.
        hidden_channels
            Number of hidden channels.
            If a float is provided, it is interpreted as a percentage of the input channels.
        layers
            The number of layers. Default: 2.
        dropout
            Dropout probability. Default: 0.0,
            If a tuple is provided, it is interpreted as (intra, post).
        norm
            Normalization layer to use.
            If `None`, no normalization is applied.
            If a tuple is provided, it is interpreted as (intra, post) or (antre, intra, post).
        bias
            Whether to add a bias term.
            If two values are provided, the final value is for the last layer.
            If `None`, a bias is added if no normalization is applied.
        activation
            Activation function to use. Default: `nn.GELU`.
            If a tuple is provided, it is interpreted as (intra, post) or (ante, intra, post).
            If `None`, no activation is applied.
        init_gain
            The gain value for the initialization of the weights. If `None`, no additional initialization is applied.
        """
        super().__init__()

        assert init_gain is None or init_gain >= 0.0, init_gain
        assert layers > 0

        # Handle hidden channels (if float, interpret as percentage of input channels)
        if isinstance(hidden_features, float):
            hidden_features = int(in_features * hidden_features)
        elif isinstance(hidden_features, int):
            pass
        else:
            msg = f"Invalid type for `hidden_channels`: {type(hidden_features)}"
            raise ValueError(msg)

        # Set attributes
        self.in_features = in_features
        self.out_features = out_features
        self.hidden_features = hidden_features

        # Handle variable parameters
        ante_norm, intra_norm, post_norm = _decompose_ante_intra_post(norm)
        ante_activation, intra_activation, post_activation = _decompose_ante_intra_post(
            activation
        )

        ante_bias, intra_bias, post_bias = _decompose_ante_intra_post(bias)
        if ante_bias is not None:
            msg = f"Ante bias is not supported. Got {bias=}"
            raise NotImplementedError(msg)
        ante_dropout, intra_dropout, post_dropout = _decompose_ante_intra_post(
            dropout, default_ante=None, default_post=dropout
        )
        if ante_dropout is not None:
            msg = f"Ante dropout is not supported. Got {dropout=}"
            raise NotImplementedError(msg)

        # Construct layers
        layer_modules: list[MLPLayer] = []
        for n in range(layers):
            is_final = n >= (layers - 1)
            n_norm = post_norm if is_final else intra_norm
            n_activation = post_activation if is_final else intra_activation
            n_dropout = post_dropout if is_final else intra_dropout
            n_bias = post_bias if is_final else intra_bias
            n_init_gain = (is_final and init_gain) or (n == 0 and 1.0) or None

            layer_modules.append(
                MLPLayer(
                    in_features if n == 0 else hidden_features,
                    out_features if is_final else hidden_features,
                    bias=n_bias,
                    norm=n_norm,
                    activation=n_activation,
                    dropout=n_dropout,
                    init_gain=n_init_gain,
                )
            )

        # Register modules
        if ante_norm is not None:
            self.input_norm = get_norm(ante_norm, in_features)
        else:
            self.register_module("input_norm", None)
        if ante_activation is not None:
            self.input_activation = get_activation(ante_activation)
        else:
            self.register_module("input_activation", None)
        assert len(layer_modules) == layers, (len(layer_modules), layers)
        self.layers = nn.Sequential(*layer_modules)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Reset the parameters of the MLP."""
        for layer in (self.input_activation, self.input_norm, *self.layers):
            with contextlib.suppress(AttributeError, NotImplementedError):
                layer.reset_parameters()

    @T.override
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the MLP.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        if self.input_norm is not None:
            x = self.input_norm(x)
        if self.input_activation is not None:
            x = self.input_activation(x)
        return self.layers(x)


class MLPLayer(nn.Module):
    """See :class:`MLP` for more details."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        *,
        bias: bool | None = None,
        norm: NormSpec = None,
        activation: ActivationSpec | None,
        dropout: float,
        init_gain: float | None,
    ) -> None:
        super().__init__()

        self.init_gain = init_gain

        norm = get_norm(norm, out_features)
        self.linear = nn.Linear(
            in_features, out_features, bias=get_default_bias(bias, norm)
        )
        self.norm = norm

        activation = get_activation(activation)
        if activation is None:
            self.register_module("activation", None)
        else:
            self.activation = activation

        self.dropout = nn.Dropout(dropout)

    def reset_parameters(self) -> None:
        for layer in (self.norm, self.activation, self.dropout):
            with contextlib.suppress(AttributeError, NotImplementedError):
                layer.reset_parameters()

        # When init_gain is specified, the user requests an initialization where the
        # output of the layer is roughly equal to the input scaled by the gain.
        if self.activation is None or self.init_gain is not None:
            nn.init.xavier_uniform_(self.linear.weight, gain=self.init_gain or 1.0)
        if self.activation is not None:
            nn.init.kaiming_uniform_(self.linear.weight, a=1, nonlinearity="leaky_relu")

        # Zero out bias if present
        if self.linear.bias is not None:
            nn.init.zeros_(self.linear.bias)

    @T.override
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear(x)
        if self.norm is not None:
            x = self.norm(x)
        if self.activation is not None:
            x = self.activation(x)
        return self.dropout(x)


_Ti = T.TypeVar("_Ti", covariant=True)
_Ta = T.TypeVar("_Ta", covariant=True)
_Tp = T.TypeVar("_Tp", covariant=True)


def _decompose_ante_intra_post(
    arg: T.Sequence[_Ta | _Ti | _Tp] | tuple[_Ti, _Ta, _Tp] | tuple[_Ti, _Tp] | _Ti,
    *,
    default_ante: _Ti = None,
    default_post: _Tp = None,
) -> tuple[_Ta | _Ti, _Ti, _Tp | _Ti]:
    """Quick macro for handling the decomposition of antre- intera- and pre- arguments."""
    if isinstance(arg, T.Iterable):
        *ante_intra, post = arg
        if len(ante_intra) == 1:
            intra = ante_intra[0]
            ante = default_ante
        else:
            ante, intra = ante_intra
    else:
        intra = arg
        ante = default_ante
        post = default_post

    return ante, intra, post  # type: ignore
